/**
 * @Copyright © 2002-2020 Acronis International GmbH. All rights reserved
 */

require([
    "app/common",
    "app/messages",
    "app/quotas",
    "app/tags",
    "app/wizardData",
    "aps/load",
    "aps/xhr",
    "aps/Button",
    "aps/Gauge",
    "aps/ResourceStore",
    "dojo/when",
    "dojo/promise/all",
    "dijit/registry"
], function (
    common,
    messages,
    quotas,
    tags,
    wizardData,
    load,
    xhr,
    Button,
    Gauge,
    Store,
    when,
    all,
    registry
) {
    var globalId = aps.context.vars.globals.aps.id,
        tenantId = aps.context.vars.tenant.aps.id,
        accountStore = new Store({
            apsType: common.TYPES.USERS,
            target: "/aps/2/resources"
        }),
        subscriptionAccountStore = new Store({
            target: "/aps/2/resources/" + tenantId + "/users",
            idProperty: "aps.id"
        }),
        subscriptionStore = new Store({
            target: "/aps/2/resources/" + tenantId + "/subscription",
            idProperty: "aps.id"
        }),
        getStatisticsDeffer = xhr.get("/aps/2/resources/" + tenantId + "/updateStatistics").then(
            function (statistics) {
                return statistics;
            }, messages.displayError),
        deleteInProgress = false,
        userViewViewId = common.getRelatedViewId('view'),
        userNewSelectViewId = common.getRelatedViewId('new-select'),
        userNewConfigureViewId = common.getRelatedViewId('new-configure'),
        userActivateViewId = common.getRelatedViewId('activate');

    function output(html) {
        return ["aps/Output", {id: common.genId('output'), value: html, escapeHTML: false}];
    }

    function image(src) {
        return "<img src=" + src + ">";
    }

    function span(text, cssClass, title) {
        return "<span" + (title ? " title='" + title + "'" : "") + (cssClass ? " class=" + cssClass : "") + ">" +
            text + "</span>";
    }

    function formatStorageUsage(usageObject) {
        var result = (usageObject.usage == null) ? '0' : common.formatStorageSize(usageObject.usage);
        if (usageObject.limit != null && usageObject.limit > 0) {
            result = result + ' / ' + common.formatStorageSize(usageObject.limit);
        }
        return result;
    }

    function formatUnitUsage(usageObject) {
        var result = (usageObject.usage == null) ? '0' : usageObject.usage;
        if (usageObject.limit != null && usageObject.limit > 0) {
            result = result + ' / ' + usageObject.limit;
        }
        return result;
    }

    function prepareUsageStats(value, limit) {
        value = Number(value);
        limit = Number(limit);
        var usageStats = (value > 0) ? value : 0;
        if (limit >= 0) {
            return usageStats + ' / ' + limit;
        }
        return usageStats.toString();
    }

    function checkButtonState() {
        var selectedRecordsCount = registry.byId("account-grid").get("selectionArray").length;
        registry.byId("newuser").set('disabled', deleteInProgress || selectedRecordsCount > 0);
        registry.byId("deleteuser").set('disabled', deleteInProgress || selectedRecordsCount === 0);
    }

    function onNewAccountRequestFail(err) {
        messages.displayError(err);
        registry.byId("newuser").cancel();
    }

    xhr.get("/aps/2/resources/" + globalId + "/settings").then(function (settings) {
        var userStore = new Store({
            apsType: !settings.ccpv1_show_staff_members ?
                "http://aps-standard.org/types/core/service-user/1.0" :
                "http://parallels.com/aps/types/pa/user/1.0",
            target: "/aps/2/resources"
        });

        all({
            statistics: getStatisticsDeffer,
            users: userStore.query()
        }).then(function (response) {
            var statistics = response.statistics,
                users = response.users,
                needRefresh = false, // if any account is in-progress we need to refresh the grid
                createBackupAccount = function () {
                    when(subscriptionStore.query(), function (subscriptions) {
                        var subscription = subscriptions[0],
                            resourcesStore = new Store({
                                target: "/aps/2/resources/" + subscription.aps.id + "/resources",
                                idProperty: "aps.id"
                            });

                        when(resourcesStore.query(), function (resources) {
                            for (var i = 0; i < resources.length; i++) {
                                if (resources[i].apsType == common.TYPES.USERS) {
                                    if (resources[i].usage >= resources[i].limit) {
                                        messages.displayWarning(_("Limit for service users has been reached."));
                                        registry.byId("newuser").cancel();
                                        return;
                                    }
                                }
                            }

                            when(accountStore.query(), function (accounts) {
                                for (var i = 0; i < users.length; i++) {
                                    var found = false;
                                    for (var j = 0; j < accounts.length; j++) {
                                        if (accounts[j].serviceUserId == users[i].aps.id) {
                                            found = true;
                                        }
                                    }
                                }

                                wizardData.resetWizardData(userNewSelectViewId);
                                wizardData.resetWizardData(userNewConfigureViewId);
                                aps.apsc.gotoView(userNewSelectViewId);
                            }, onNewAccountRequestFail);
                        }, onNewAccountRequestFail);
                    }, onNewAccountRequestFail);
                },
                deleteBackupAccount = function () {
                    if (!confirm(_("Are you sure that you want to unassign this service from the user(s)? All user backups will be deleted."))) {
                        this.cancel();
                        return;
                    }

                    var accountGrid = registry.byId("account-grid"),
                        selectionArray = accountGrid.get("selectionArray"),
                        deffers = [],
                        deletedIds = [],
                        deleteBtn = this;

                    messages.removeAll();

                    selectionArray.forEach(function (id) {
                        var deffer = subscriptionAccountStore.remove(id).then(function () {
                            deletedIds.push(id);
                        }, messages.displayErrorAndGo);
                        deffers.push(deffer);
                    });

                    deleteInProgress = true;
                    checkButtonState();

                    all(deffers).then(function () {
                        accountGrid.refresh();

                        deletedIds.forEach(function (id) {
                            selectionArray.splice(selectionArray.indexOf(id), 1);
                        });

                        deleteInProgress = false;
                        deleteBtn.cancel();
                        checkButtonState();
                    });
                },
                toolbar = ["aps/Toolbar", [
                    ["aps/ToolbarButton", {
                        id: "newuser",
                        label: _("Add New Service User"),
                        iconClass: "sb-add-new",
                        onClick: createBackupAccount
                    }],
                    ["aps/ToolbarButton", {
                        id: "deleteuser",
                        label: _("Delete"),
                        iconClass: "sb-delete",
                        onClick: deleteBackupAccount
                    }]
                ]],
                getStorageCell = function () {
                    for (let counter of common.getAvailableCountersForArea('storage_usage_only')) {
                        var firstUserUsage = Object.values(statistics.userUsage)[0];

                        if (!firstUserUsage || !firstUserUsage[counter.name]) {
                            continue;
                        }

                        return {
                            name: span(tags.icon(counter.iconClass + ' acronis-gray-icon', 16) + quotas.formatStorageLabel(counter.label, null),
                                'aconis-ccp1-storage-usage-th-title', quotas.formatStorageLabel(counter.label, null)),
                            escapeHTML: false,
                            sortable: false,
                            renderCell: function (row) {
                                var usage = statistics.userUsage[row.userId];
                                if (!usage) {
                                    return _("N/A");
                                }

                                var data = usage[counter.name];
                                if (data.limit == common.QUOTA_UNLIMITED_VALUE) {
                                    if (data.usage == null) {
                                        return _("N/A");
                                    }
                                    return common.formatStorageSize(data.usage);
                                }

                                var classesMap = {};
                                classesMap[Math.round(data.limit * 0.5)] = "warn";
                                classesMap[Math.round(data.limit * 0.75)] = "over";
                                var currentUsageForGauge = (data.usage == null) ? _("N/A") : common.formatStorageSize(
                                    data.usage);
                                return new Gauge({
                                    label: counter.label,
                                    legend: currentUsageForGauge + " " + _("used of") + " " + common.formatStorageSize(
                                        data.limit),
                                    minimum: 0,
                                    maximum: data.limit,
                                    value: data.usage == null ? 0 : data.usage,
                                    classesMap: classesMap
                                });
                            }
                        };
                    }

                    // Return empty cell if no storage counter of any edition was found
                    return {
                        name: span(tags.icon('i-cloud-storage-o--16 acronis-gray-icon', 16) + quotas.formatStorageLabel(_('Cloud Storage {unit}'), null),
                            'aconis-ccp1-storage-usage-th-title', quotas.formatStorageLabel(_('Cloud Storage {unit}'), null)),
                        escapeHTML: false,
                        sortable: false,
                        renderCell: function () {
                            return _("N/A");
                        }
                    };
                },
                gridColumns = [
                    {
                        name: image("images/user_statuses/ico_state_on_16px.png"),
                        field: "userstatus",
                        escapeHTML: false,
                        renderCell: function (row, data) {
                            switch (data) {
                                case "1":
                                    return image("images/user_statuses/ico_state_on_16px.png");
                                case "2":
                                    needRefresh = true;
                                    return image("images/user_statuses/wheel-progress.gif");
                                default:
                                    return image("images/user_statuses/ico_state_off_16px.png");
                            }
                        }
                    },
                    {name: _("Login"), field: "endUserLogin", type: "resourceName"},
                    {
                        name: _("Name"),
                        field: "displayName",
                        renderCell: function (row) {
                            for (var i = 0; i < users.length; i++) {
                                if (row.serviceUserId == users[i].aps.id) {
                                    return common.escapeAngularBrackets(users[i].displayName);
                                }
                            }
                            return "";
                        }
                    },
                    getStorageCell()
                ],
                userGrid = ["aps/Grid", {
                    id: "account-grid",
                    showPaging: false,
                    recordsTotalText: _("item(s) total"),
                    columns: gridColumns,
                    store: subscriptionAccountStore,
                    selectionMode: 'multiple',
                    apsResourceViewId: userViewViewId
                }],
                totalItems = [],
                visibleQuotasTotalList = common.getTenantVisibleQuotas(statistics.tenant, '', false),
                visibleQuotasRowList = common.getTenantVisibleQuotas(statistics.tenant, 'quota_controls_ccpv1_users_list', false);

            for (var [counterName, counterData] of Object.entries(common.getCountersToOfferingItems())) {
                if (visibleQuotasTotalList[counterName]) {
                    var formattedLimit,
                        counterEditionName = visibleQuotasTotalList[counterName];
                    switch (counterData.legendSchema) {
                        case "feature":
                            formattedLimit = span(_('Enabled'));
                            break;
                        case "storage":
                            formattedLimit = span(formatStorageUsage(statistics.tenant[counterEditionName]));
                            break;
                        default:
                            formattedLimit = span(formatUnitUsage(statistics.tenant[counterEditionName]));
                    }
                    totalItems.push(span(
                        tags.icon(counterData.iconClass + ' acronis-gray-icon', 16) + formattedLimit,
                        'acronis-total-usage',
                        quotas.formatStorageLabel(counterData.label, null)
                    ));
                } else {
                    totalItems.push('');
                }
            }

            let counter = common.getCounterData('CloudStorage');
            if (counter) {
                gridColumns.push({
                    name: span(tags.icon(counter.iconClass + ' acronis-gray-icon', 16), null, counter.label),
                    escapeHTML: false,
                    sortable: false,
                    renderCell: function (row) {
                        let usage = statistics.userUsage[row.userId];

                        if (!usage) {
                            return _("N/A");
                        }
                        let data = usage[counterName];

                        if (!data) {
                            return _("N/A");
                        }

                        return prepareUsageStats(data.usage, data.limit);
                    }
                });
            }


            gridColumns.push({
                name: "",
                renderCell: function (user) {
                    switch (user.userstatus) {
                        case common.USER_STATUS.ACTIVE:
                            return new Button({
                                id: common.genId("login_button"),
                                label: _("Log into Backup Console"),
                                'class': 'acronis-ccp1-button log-into-backup-console',
                                iconClass: "sb-export",
                                onClick: function () {
                                    var self = this,
                                        w = window.open('', '_blank');

                                    common.loginToConsole(user).then(function (uri) {
                                        w.location.href = uri;
                                        self.cancel();
                                    }, function (error) {
                                        w.close();
                                        console.error("ERR: \n" + JSON.stringify(error));
                                        messages.displayError(error);
                                        self.cancel();
                                    });

                                }
                            });
                        case common.USER_STATUS.NOT_ACTIVATED:
                            return new Button({
                                id: common.genId("activate_button"),
                                label: _("Activate"),
                                'class': 'acronis-ccp1-button log-into-backup-console',
                                onClick: function () {
                                    aps.apsc.gotoView(userActivateViewId, user.aps.id);
                                }
                            });
                    }
                    return "";
                }
            });

            load(["aps/PageContainer", {id: "apsPageContainer", 'class': "acronis-tenant"}, [
                ['aps/Output', {
                    value: _("A unique integration of backup with full-stack next-generation anti-malware protection and comprehensive endpoint management tools. This synergy eliminates complexity and improves security so you can have better protection for your data.") + tags.br() + tags.br(),
                    escapeHTML: false
                }],
                ["aps/Container", {
                    id: common.genId('container'),
                    title: _("Service users"),
                    description: _("A service user can access the management console and manage backups of his/her own data.")
                }, [
                    toolbar, userGrid
                ]]
            ]]).then(function () {
                if (statistics.error) {
                    messages.displayMessage(statistics.error, 'error');
                }
                (function setRefreshTimeout() {
                    setTimeout(function () {
                        xhr.get("/aps/2/resources/" + tenantId + "/updateStatistics").then(
                            function (newStatistics) {
                                statistics = newStatistics;
                                registry.byId("account-grid").refresh();
                                needRefresh = false;
                            }, messages.displayError);
                    }, 5000);
                })();

                checkButtonState();
                registry.byId("account-grid").get("selectionArray").watchElements(checkButtonState);
            });
        });
    }, messages.displayError);
});


