/**
 * @Copyright © 2002-2020 Acronis International GmbH. All rights reserved
 */

define([
    "dojo/_base/declare",
    "dojo/_base/lang",
    "dojo/Deferred",
    "dojo/promise/all",
    "dijit/registry",
    "dojox/mvc/at",
    "dojox/mvc/getStateful",
    "dojox/mvc/getPlainValue",
    "aps/_View",
    "aps/ResourceStore",
    "aps/Container",
    "aps/Memory",
    "aps/load",
    "aps/xhr",
    "./../utils",
    "./../components/getUserResourcesGrid",
    "./../../common",
    "./../../messages"
], function (
    declare,
    lang,
    Deferred,
    all,
    registry,
    at,
    getStateful,
    getPlainValue,
    _View,
    Store,
    Container,
    Memory,
    load,
    xhr,
    utils,
    getUserResourcesGrid,
    common,
    messages
) {
    function getBackupAccount(tenant, user) {
        var deferred = new Deferred();
        xhr.get("/aps/2/resources/" + tenant.aps.id + "/users?eq(user.aps.id," + user.aps.id + ")").then(function (users) {
            if (users && users.length) {
                common.getBackupUserData(users.pop()).then(function(user) {
                    deferred.resolve(user);
                })
            } else {
                deferred.resolve(null);
            }
        }, deferred.reject);
        return deferred;
    }

    function createModelForNewBackupAccount(tenant, user, profile, needActivation, countersList) {
        let userQuota = common.getUserQuota(profile),
            tenantQuotas = Object.values(common.getTenantVisibleQuotas(tenant, '', false)),
            account = {
                aps: {
                    type: common.TYPES.USERS,
                    subscription: tenant.aps.subscription
                },
                user: [{aps: {id: user.aps.id}}],
                tenant: [{aps: {id: tenant.aps.id}}],

                endUserLogin: user.login,
                endUserPassword: null,
                activateDuringCreation: needActivation,

                notifications: common.transformNotificationsToObject(profile.notifications),

                privileged: common.USER_PRIVILEGED.NO,

                offeringItems: [],
                userUsage: {}
            };
        countersList.forEach(function (counter) {
            if (tenantQuotas.includes(counter.name)) {
                account.offeringItems.push({
                    name: window.offeringItemsMap[counter.name],
                    limit: userQuota ? 0 : common.QUOTA_UNLIMITED_VALUE
                });
            }
        });

        return getStateful(account);
    }

    function createModelForExistingBackupAccount(backupAccount, backupAccountStatistic, countersList) {
        var account = {
            aps: {
                id: backupAccount.aps.id
            },

            endUserLogin: backupAccount.endUserLogin,

            notifications: common.transformNotificationsToObject(backupAccount.notifications),

            privileged: backupAccount.privileged,

            offeringItems: backupAccount.offeringItems,
            userUsage: {}
        };

        if (backupAccountStatistic && backupAccountStatistic.userUsage[backupAccount.userId]) {
            let usage = backupAccountStatistic.userUsage[backupAccount.userId];
            countersList.forEach(function (counter) {
                if (usage[counter.name]) {
                    account.userUsage[counter.name] = usage[counter.name].usage;
                }
            });
        }
        return getStateful(account);
    }

    function submitDeltas(aps, deltas, subscription) {
        if (deltas.length) {
            return aps.biz.requestUsageChange({
                deltas: deltas,
                operations: [],
                subscriptionId: subscription
            }).then(function () {
                return aps.biz.commit();
            });
        } else {
            var deferred = new Deferred();
            deferred.resolve();
            return deferred;
        }
    }

    return declare(_View, {
        _container: null,

        _user: null,
        _tenant: null,
        _model: null,
        _profile: null,

        _isNewUser: true,
        _needActivation: false,

        _deltas: [],

        _submit: function () {
            var self = this;
            var actions = {};

            if (Object.keys(self._model.notifications).length) {
                self._model.notifications = common.transformNotificationsToArray(self._model.notifications);
            } else {
                delete self._model.notifications;
            }
            delete self._model.userUsage;

            if (self._model.aps.id) {
                actions.configureResource = [getPlainValue(self._model)];
            }
            else {
                actions.provisionResource = [getPlainValue(self._model)];
            }

            var operation = {
                'call': {
                    'resourceId': self._tenant.aps.id,
                    'operation': 'setUsersLicensesAfterUpsell',
                    'parameters': {'data': actions}
                }
            };
            var subscription = self._tenant.aps.subscription;

            // request is split into two due to internal limitations and bad design of APS
            return submitDeltas(aps, self._deltas, subscription)
            .then(function () {
                return aps.biz.requestUsageChange({
                    deltas: [],
                    operations: [operation],
                    subscriptionId: subscription
                }).then(function () {
                    return aps.biz.commit();
                }).then(function () {
                    self.nav.gotoView(common.APP_ID + (self._tenant.type == common.TENANT_TYPE.PARTNER ? "#nextcp-ccp-partner" : "#nextcp-ccp"));
                });
            });
        },
        _renderForm: function () {
            var self = this, form = [];

            self._container.removeAll();

            var fields = [];
            if (self._isNewUser) {
                fields.push(['aps/TextBox', {
                    id: self.genId('login'),
                    label: _('Login'),
                    value: at(self._model, 'endUserLogin'),
                    gridSize: 'md-4 xs-12',
                    required: true,
                    extraValidator: (function () {
                        var result = false,
                            previousValue = '';

                        return function (value) {
                            if (previousValue === value) {
                                return result;
                            }

                            var login = this;
                            xhr.get('/aps/2/resources/' + self._tenant.aps.id + '/loginExists', {
                                query: {login: value, datacenterId: ''},
                                sync: true
                            }).then(function (existing) {
                                if (existing) {
                                    login.set('invalidMessage', _('This login already exists.'));
                                }
                                result = !existing;
                                previousValue = value;
                            }, messages.displayError);

                            return result;
                        };
                    })()
                }]);

                if (self._needActivation) {
                    fields.push(['aps/Password', {
                        id: self.genId('password'),
                        label: _('Password'),
                        value: at(self._model, 'endUserPassword'),
                        gridSize: 'md-4 xs-12',
                        required: true,

                        showResetButton: true,
                        passwordVisible: true,
                        showStrengthIndicator: true,
                        constraints: {
                            minimumLength: 8,
                            maximumLength: 16
                        },
                        generator: utils.passwordGenerator,
                        extraValidator: function (value) {
                            var res = utils.validatePassword(value);
                            if (res === true) {
                                return true;
                            }
                            else {
                                this.set('invalidMessage', res);
                                return false;
                            }
                        }
                    }]);
                }

                form.push(['aps/Panel', {title: _('Set service user credentials')}, [
                    ['aps/FieldSet', {gridSize: 'md-12 xs-12'}, fields]
                ]]);
            }

            form.push(['aps/Tiles', {title: _('Service User Settings')}, [
                ['aps/Tile',{title: _('Management Portal')},  [
                    ['aps/CheckBox', {
                        description: _('Quota overuse notification'),
                        checked: at(self._model.notifications, 'quota')
                    }],
                    ['aps/CheckBox', {
                        description: _('Schedule usage report'),
                        checked: at(self._model.notifications, 'reports')
                    }]
                ]],
                ['aps/Tile', {title: _('Cyber Protection')}, [
                    ['aps/CheckBox', {
                        description: _('Failure notifications'),
                        checked: at(self._model.notifications, 'backup_error')
                    }],
                    ['aps/CheckBox', {
                        description: _('Warning notifications'),
                        checked: at(self._model.notifications, 'backup_warning')
                    }],
                    ['aps/CheckBox', {
                        description: _('Success notifications'),
                        checked: at(self._model.notifications, 'backup_info')
                    }],
                    ['aps/CheckBox', {
                        description: _('Daily recap about active alerts'),
                        checked: at(self._model.notifications, 'backup_daily_report')
                    }]
                ]],
                ['aps/Tile', {title: _('Privileges')}, [
                    ['aps/FieldSet', [
                        ['aps/CheckBox', {
                            description: _('Administrator'),
                            checked: at(self._model, 'privileged').transform({
                                parse: function(value) {
                                    return value ? common.USER_PRIVILEGED.YES : common.USER_PRIVILEGED.NO;
                                },
                                format: function(value) {
                                    return value === common.USER_PRIVILEGED.YES;
                                }
                            })
                        }],
                        ['aps/Output', {description: _('An administrator can manage services of other users and back up Microsoft 365 seats.')}]
                    ]]
                ]]
            ]]);

            load(form, self._container).then(function () {
                if (self._needActivation) {
                    var password = registry.byId(self.genId('password'));
                    if (password) {
                        password.generate();
                    }
                }
                if (common.getUserQuota(self._profile)) {
                    getUserResourcesGrid(self._tenant, self._model, 1).then(function (resources) {
                        self._container.addChild(resources.grid);
                        self._deltas = resources.deltas;
                        aps.apsc.hideLoading();
                    }, messages.displayError);
                }
                else {
                    aps.apsc.hideLoading();
                }
            }, messages.displayError);
        },
        init: function () {
            this._container = new Container().placeAt(this.domNode);
        },
        onHide: function () {
            var self = this;
            self._container.removeAll();
        },
        onContext: function (context) {
            var self = this,
                selectedUserId = aps.context.params && aps.context.params.userId,
                currentUserId = aps.context.user.aps.id;

            if (!selectedUserId) {
                return;
            }
            self._isNewUser = !!aps.context.params.isNewUser;
            self._needActivation = self._isNewUser && selectedUserId === currentUserId;

            self._tenant = context.vars.tenant;

            var label = self._isNewUser ? _("Service Users Settings") : _('Change Service User Settings');
            self.set('label', label);

            all({
                user: xhr.get('/aps/2/resources/' + selectedUserId),
                profile: common.getBackupProfile(self._tenant)
            }).then(function (response) {
                let countersList = common.getAvailableCountersForArea('user_quotas');
                self._user = response.user;
                self._profile = response.profile;
                if (self._isNewUser) {
                    self._model = createModelForNewBackupAccount(self._tenant, self._user, self._profile, self._needActivation, countersList);
                    self._renderForm();
                    return;
                }
                getBackupAccount(self._tenant, self._user).then(function (backupAccount) {
                    if (!backupAccount) {
                        messages.removeAll();
                        messages.displayMessage(_('Backup account does not exists for user __displayName__.', self._user), 'error');
                        return;
                    }
                    xhr.get('/aps/2/resources/' +  self._tenant.aps.id  + '/updateStatistics').then(function (backupAccountStatistic) {
                        self._model = createModelForExistingBackupAccount(backupAccount, backupAccountStatistic, countersList);
                        self._renderForm();
                    }, messages.displayError);
                }, messages.displayError);
            }, messages.displayError);
        },
        onSubmit: function () {
            var self = this;
            if (!self.validate()) {
                aps.apsc.cancelProcessing();
                return;
            }
            self._submit().then(function () {
                aps.apsc.prev();
            });
        },
        onCancel: function () {
            aps.apsc.prev();
        }
    });
});




