/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

require([
    "aps/load",
    "aps/Memory",
    "aps/ResourceStore",
    "aps/xhr",
    "dijit/registry",
    "dojox/mvc/at",
    "dojox/mvc/getPlainValue",
    "dojox/mvc/getStateful",
    "./js/common.js",
    "./js/messages.js"
], function (
    load,
    Memory,
    Store,
    xhr,
    registry,
    at,
    getPlainValue,
    getStateful,
    common,
    messages
) {
    function isUniqueCurrency(commitments) {
        var distinct = {};
        return !commitments.some(function (commitment) {
            var currency = commitment.currency.toUpperCase();
            if (distinct[currency]) {
                return true;
            }
            distinct[currency] = true;
            return false;
        });
    }

    var globals = aps.context.vars.globals,
        tiersStore = new Store({
            target: "/aps/2/resources/" + globals.aps.id + "/minimalCommitmentTiers"
        }),
        tierModel = getStateful({
            aps: {type: common.TYPES.MINIMAL_COMMITMENT_TIER},
            name: "",
            level: -1,
            minimalCommitments: []
        });

    var levels = [
        {value: '-1', label: _("No commitment")},
        {value: '0', label: _("Minimum")},
        {value: '1', label: "1"},
        {value: '2', label: "2"},
        {value: '3', label: "3"},
        {value: '4', label: "4"},
        {value: '5', label: "5"},
        {value: '6', label: "6"},
        {value: '7', label: "7"}
    ];
    var widgets = ["aps/PageContainer", {id: "apsPageContainer"}, [
        ["aps/FieldSet", {title: _("General")}, [
            ["aps/TextBox", {
                id: "name",
                label: _("Minimum commitment tier name"),
                value: at(tierModel, "name"),
                required: true,
                trim: true
            }],
            ["aps/Select", {
                id: "level",
                label: _("Level"),
                value: at(tierModel, "level").transform({
                    format: function (value) {
                        return String(value);
                    },
                    parse: function (value) {
                        return parseInt(value, 10);
                    }
                }),
                required: true,
                store: new Memory({data: levels, idProperty: "value"})
            }]
        ]],
        ["aps/WidgetList", {
            children: at(tierModel, "minimalCommitments"),
            visible: at(tierModel, "level").transform({
                format: function (value) {
                    return (0 <= parseInt(value, 10));
                }
            }).direction(at.from)
        }, [
            ["aps/FieldSet", {title: _("Minimum commitment")}, [
                ["aps/TextBox", {
                    label: _("Currency"),
                    value: at("rel:", "currency").transform({
                        format: function (value) {
                            return String(value).toUpperCase();
                        },
                        parse: function (value) {
                            return String(value).toUpperCase();
                        }
                    }),
                    required: true,
                    pattern: '^[A-Za-z]{3}$',
                    hint: _("For example, USD")
                }],
                ["aps/TextBox", {
                    label: _("Value"),
                    value: at("rel:", "value"),
                    required: true,
                    pattern: '^(0|[1-9]\\d*)(\\.\\d{1,2})?$',
                    hint: _("For example, 123.45")
                }]
            ]]
        ]]
    ]];

    load(widgets).then(function () {
        aps.app.onSubmit = function () {
            var page = registry.byId("apsPageContainer");
            if (!page.validate()) {
                aps.apsc.cancelProcessing();
                return;
            }
            var tierData = getPlainValue(tierModel);
            if (tierData.level < 0) {
                tierData.minimalCommitments = [];
            }
            if (!isUniqueCurrency(tierData.minimalCommitments)) {
                messages.displayWarning(_('Please remove values of minimum commitment in the same currency.'), false);
                aps.apsc.cancelProcessing();
                return;
            }
            tiersStore.put(tierData).then(function () {
                aps.apsc.gotoView(common.getRelatedViewId('list'));
            }, messages.displayError);
        };
    }, messages.displayError);


    aps.app.onCancel = function () {
        aps.apsc.gotoView(common.getRelatedViewId('list'));
    };
});
