<?php
/**
 * @Copyright © 2002-2020 Acronis International GmbH. All rights reserved
 */

use Acronis\App\Base\ApplicationsManager;
use Acronis\App\CloudApi\ApiInterface;
use Acronis\App\Services\CountersManager\CountersManagerAwareTrait;
use Model_PropertyConverters_NotificationConverter as NotificationConverter;
use Model_PropertyConverters_ServiceRolesConverter as ServiceRolesConverter;

class Model_PropertyConverters_UsersConverterHelper
{
    use CountersManagerAwareTrait;

    private $convertOfferingItems = false;

    private $method;

    private $defaultNotifications = [];

    private $defaultApplications = [];

    public function __construct($method)
    {
        $this->method = $method;
    }

    /**
     * Converts old user properties from $resource to new properties inside $destination
     * @param users|object $resource
     * @param users $destination
     */
    public function convert($resource, $destination)
    {
        if ($this->convertOfferingItems) {
            $this->convertUserOfferingItems($resource, $destination);
        }

        $this->convertUserNotifications($resource, $destination);
        $this->convertUserServiceRoles($resource, $destination);
    }

    /**
     * @param $isEnabled
     * @param $isActivated
     * @return int
     */
    public static function getUserStatus($isEnabled, $isActivated)
    {
//-----------------------------------------------------------------
//| Enabled in Cloud | Active in Cloud | Status in APIv1 response |
//|------------------|-----------------|--------------------------|
//|      TRUE        |      TRUE       |           1              |
//|------------------|-----------------|--------------------------|
//|      TRUE        |      FALSE      |          -1              |
//|------------------|-----------------|--------------------------|
//|      FALSE       |      TRUE       |           0              |
//|------------------|-----------------|--------------------------|
//|      FALSE       |      FALSE      |           0              |
//|------------------|-----------------|--------------------------|
        if (!$isEnabled) {
            return users::USER_STATUS_INACTIVE;
        }

        return $isActivated ? users::USER_STATUS_ACTIVE : users::USER_STATUS_ERROR;
    }

    /**
     * Collects limits from counters and coverts into list of corresponding OfferingItems,
     * should be removed till APS3.0
     * @see convertProperties
     * @param users|object $resource
     * @param users $destination
     * @deprecated
     */
    private function convertUserOfferingItems($resource, $destination)
    {
        $offeringItems = [];
        $edition = $this->fetchSubscriptionEdition($destination->tenant, $this->getCountersManager());
        foreach (Model_Resource_Tenant::$userQuotasPropertiesMap as $tenantCounterName => $userCounterName) {
            $tenantCounterName = $this->addLegacyEditionPrefix($tenantCounterName, $edition);
            $counter = $this->getCountersManager()->getCounterByName($tenantCounterName);
            $offeringItem = new UserOfferingItem();
            $offeringItem->name = $counter->getOfferingItemName();
            $limit = Arr::getNestedItem($resource, [$userCounterName, 'limit'], 0);
            if ($limit != 0) {
                if ($limit == Model_MspGroupManager::UNLIMITED_QUOTA_VALUE) {
                    $limit = ApiInterface::UNLIMITED_OFFERING_ITEM_VALUE;
                }
                $offeringItem->limit = $limit;

                $offeringItems[] = $offeringItem;
            }
        }

        $destination->offeringItems = $offeringItems;
    }

    /**
     * @param $tenant
     * @param $countersManager
     * @return string
     */
    private function fetchSubscriptionEdition($tenant, $countersManager)
    {
        $resources = $tenant->subscription->resources();
        foreach ($resources as $resource) {
            if (property_exists($resource, 'property')) {
                $counter = $countersManager->getCounterByName($resource->property);
                if ($counter && $counter->getAggregationRules()) {
                    return $counter->getEdition();
                }
            }
        }

        return '';
    }

    /**
     * @param $stdCounterName
     * @param $edition
     * @return string
     */
    private function addLegacyEditionPrefix($stdCounterName, $edition)
    {
        switch ($edition) {
            case 'advanced':
                return 'Adv' . $stdCounterName;
            case 'cyber_protect_adv':
                return 'PAdv' . $stdCounterName;
            case 'disaster_recovery':
                return 'Dre' . $stdCounterName;
            case 'cyber_protect_dre':
                return 'PDre' . $stdCounterName;
            case 'cyber_protect_std':
                return 'P' . $stdCounterName;
            default:
                return $stdCounterName;
        }
    }

    /**
     * Converts bit mask notifyBackup field and boolean flag notificationsManagement
     * into string list of notifications supported by APIv2,
     * should be removed till APS3.0 along with NotificationConverter
     * @see convertProperties
     * @param users|object $resource
     * @param users $destination
     */
    private function convertUserNotifications($resource, $destination)
    {
        switch ($this->method) {
            case "upgrade":
                $destination->notifications = NotificationConverter::getNotifications(
                    Arr::get($resource, 'notifyBackup', 0),
                    Arr::get($resource, 'notificationsManagement', false)
                );
                break;
            case "provision":
            default:
                if (!is_array($destination->notifications)) {
                    $destination->notifications = $this->defaultNotifications;
                }
        }
    }

    /**
     * Converts privileged field (simply 0 or 1)
     * into string list of service roles supported by APIv2,
     * should be removed till APS3.0 along with ServiceRolesConverter
     * @param users|object $resource
     * @param users $destination
     */
    private function convertUserServiceRoles($resource, $destination)
    {
        $destination->serviceRoles = ServiceRolesConverter::getRoles($resource, $this->defaultApplications);
    }

    /**
     * @param array $defaultNotifications
     */
    public function setDefaultNotifications(array $defaultNotifications)
    {
        $this->defaultNotifications = $defaultNotifications;
    }

    /**
     * @param bool $convertOfferingItems
     * @return Model_PropertyConverters_UsersConverterHelper
     */
    public function setConvertOfferingItems($convertOfferingItems)
    {
        $this->convertOfferingItems = $convertOfferingItems;
    }

    /**
     * @param array $defaultApplications
     */
    public function setDefaultApplications(array $defaultApplications)
    {
        $this->defaultApplications = $defaultApplications;
    }
}