<?php
/**
 * @Copyright © 2002-2020 Acronis International GmbH. All rights reserved
 */

use Acronis\App\CloudApi\CloudUserObject;
use Acronis\Cloud\Client\Model\Common\Contact\Contact;
use Acronis\Cloud\Client\Model\Common\Types\UserNotification;
use Acronis\Cloud\Client\Model\Users\User;
use Acronis\Cloud\Client\Model\Users\UserPost;
use Acronis\Cloud\Client\Model\Users\UserPut;

/**
 * Class Model_Resource_Tenant_Partner
 */
class Model_Resource_Tenant_Partner extends Model_Resource_Tenant
{
    //default notification values for administrator account in baas
    const DEFAULT_GROUP_ADMIN_NOTIFY_BACKUP = 7;
    const DEFAULT_GROUP_ADMIN_NOTIFY_MANAGE = 1;

    /* ======= RESOURCE METHODS ======= */

    public function provision()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->checkProfile();
        $this->linkProfile();

        $tenant = $this->tenant;
        $tenant->datacenterId = $this->resolveDatacenterId();
        $this->linkDatacenter();

        $this->updateCounterLimits(true);

        $tenant->mspGroupId = $this->_createGroupForPartner();
        $this->updateTenantId();
    }

    public function provisionAsync()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->updateStartProductionTime();

        $this->linkDatacenter(false);
        $this->linkProfile(false);

        $this->getMspGroupManager()->updatePrivilegesForCustomers($this->tenant, $this);

        $this->_subscribeToOnSubscriptionChangeEvent();
        $this->_subscribeToOnUserChangeEvent();
        $this->_subscribeToOnAccountChangeEvent();

        $admins = $this->getAdministrators();
        if (count($admins) == 0) {
            $this->createAdministratorAccount();
        }
    }

    public function retrieve()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;
        $this->linkProfile($realLink);
        $this->linkDatacenter($realLink);

        if (is_null($tenant->mspGroupId)) {
            return;
        }
        $this->updateCounterLimits();
        $this->_retrieveUsageForPartner();
    }

    /**
     * @param $new
     * @throws Exception
     */
    public function configure($new)
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        try {

            $tenant = $this->tenant;
            $realLink = !$tenant->subscription->disabled;

            $this->_updateProfileLink($realLink);
            $this->linkDatacenter($realLink);

            if (is_null($tenant->mspGroupId)) {
                return;
            }

            $this->_updateGroupForPartner();
            $this->getMspGroupManager()->updatePrivilegesForCustomers($this->tenant, $this);

        } catch (\Exception $exception) {
            $this->handleException($exception);
        }
    }

    public function unprovision()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        if ($tenant->ismigration) {
            Log::getInstance()->event(static::LOG_UNPROVISIONING_TENANT_IS_MIGRATION);
            return;
        }

        if (is_null($tenant->mspGroupId)) {
            Log::getInstance()->warning(static::LOG_WARNING_UNPROVISION_TENANT_DOES_NOT_HAVE_GROUP);
            return;
        }
        Log::getInstance()->event(sprintf(static::LOG_FORMAT_DELETE_MSP_GROUP, $tenant->account->companyName));

        $this->linkDatacenter(false);

        try {
            $this->_deleteGroupForPartner();
        } catch (Service_BaseClient_HTTPException $e) {
            if ($e->getCode() != Service_BaseClient_HTTPException::HTTP_CODE_VERSION_EXPIRED) {
                throw $e;
            }
            sleep(10);
            $this->_deleteGroupForPartner();
        }
    }

    public function upgrade()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        if (is_null($tenant->datacenterId)) {
            $tenant->datacenterId = globals::DEFAULT_DATACENTER_ID;
        }

        $this->updateStartProductionTime();
        $this->updateTenantId();
        $this->update();

        $realLink = !$tenant->subscription->disabled;
        if ($realLink) {
            $this->checkProfile();
        }

        $this->linkProfile($realLink);
        $this->linkDatacenter($realLink);

        $this->_subscribeToOnSubscriptionChangeEvent();
        $this->_subscribeToOnUserChangeEvent();
        $this->_subscribeToOnAccountChangeEvent();
    }

    /**
     * Further use in dedicated command which should be called after package upgrade  from .spec
     * @throws \Rest\RestException
     */
    public function syncCloud()
    {
        $tenant = $this->tenant;
        $this->_getOrCreateParentGroupId($tenant->datacenter);  // add parents in map
        $this->_updateGroupForPartner();
        $this->getMspGroupManager()->updatePrivilegesForCustomers($tenant, $this);
    }

    /* ======= RESOURCE EVENTS ======= */

    public function onTenantAvailable($event)
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->linkProfile();
        $this->linkDatacenter();
    }

    public function onSubscriptionChange($event)
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $needUpdate = $this->updateStartProductionTime();
        if ($this->updateCounterLimits()) {
            $needUpdate = true;
        }
        if ($needUpdate) {
            $this->update();
        }
        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;
        $this->_updateProfileLink($realLink);
        $this->linkDatacenter($realLink);
        $this->_updateGroupForPartner();
        if (!$tenant->subscription->disabled) {
            $this->getMspGroupManager()->updatePrivilegesForCustomers($this->tenant, $this);
        }
    }

    public function onUserChanged($event)
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->_updateGroupForPartner();
    }

    public function onAccountChange($event)
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->_updateGroupForPartner();
    }

    /* ======= RESOURCE OPERATIONS ======= */

    /**
     * @return array
     */
    public function getAdministrators()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        /**
         * @var Service_MspApiClient_Response_Struct_AdministratorItem[] $mspAdmins
         */
        $cloudApi = $this->createCloudApiInstance($this->tenant->datacenter);
        $cloudAdmins = $cloudApi->getTenantUsers($tenant->tenantId);
        $adminUsers = array();
        /** @var User $existingAdmin */
        foreach ($cloudAdmins as $adminId) {
            $existingAdmin = $cloudApi->getUser($adminId);
            $adminUsers[] = new CloudUserObject($existingAdmin, $this->tenant->datacenterId, $this->tenant->datacenter->name);
        }
        return $adminUsers;
    }

    /**
     * @param $adminId
     * @param string $datacenterId
     * @throws \Rest\NoContent
     */
    public function resendAdministratorLink($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $admin = $this->getMspApi()->getAdministratorById($adminId);
        $this->getMspApi()->postMailActivate(
            new Service_MspApiClient_Request_Struct_PostMailActivate($admin->getEmail(), $admin->getLogin()
            ));
        $this->throwNoContent();
    }

    /**
     * @param $adminId
     * @param string $datacenterId
     * @throws \Rest\NoContent
     */
    public function enableAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $cloudApi = $this->createCloudApiInstance($this->tenant->datacenter);
        $user = $cloudApi->getUser($adminId);

        $userUpdate = new UserPut();
        $userUpdate->setVersion($user->getVersion());

        $userUpdate->setEnabled(true);
        $cloudApi->updateUser($adminId, $userUpdate);

        $this->throwNoContent();
    }

    /**
     * @param $adminId
     * @param string $datacenterId
     * @throws \Rest\RestException
     */
    public function deleteAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $adminInfo = $this->getMspApi()->getAdministratorById($adminId);
        $this->getMspApi()->deleteAdministrator($adminId, $adminInfo->getVersion());
        $this->throwNoContent();
    }

    /**
     * @param $adminId
     * @param string $datacenterId
     * @throws \Rest\RestException
     */
    public function disableAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $cloudApi = $this->createCloudApiInstance($this->tenant->datacenter);
        $user = $cloudApi->getUser($adminId);

        $userUpdate = new UserPut();
        $userUpdate->setVersion($user->getVersion());

        if (!$user->getEnabled()) {
            $this->throwNoContent();
        }

        $userUpdate->setEnabled(false);
        $cloudApi->updateUser($adminId, $userUpdate);
        $this->throwNoContent();
    }

    /**
     * @return array
     * @throws Exception
     */
    public function getStorages()
    {
        Log::getInstance()->event(
            static::LOG_METHOD_CALL,
            $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        );

        return $this->memoize(function () {
            $response = [];
            $tenant = $this->tenant;
            $storages = $this->getMspApi()->getStorages($tenant->mspGroupId);
            if (is_null($storages)) {
                Log::getInstance()->error(static::EXCEPTION_CAN_NOT_RECEIVE_STORAGES_MESSAGE);
                throw new Exception(static::EXCEPTION_CAN_NOT_RECEIVE_STORAGES_MESSAGE);
            }

            foreach ($storages as $storage) {
                /** @var Service_MspApiClient_Response_Struct_Storage $storage */
                if ($storage->getInherited()) {
                    continue;
                }
                $response[] = [
                    static::FIELD_STORAGE_ID => (string)$storage->getId(),
                    static::FIELD_STORAGE_NAME => $storage->getName(),
                ];
            }

            return $response;
        });
    }

    /**
     * @return object
     */
    public function getDatacenter()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;
        $this->linkDatacenter($realLink);
        return (object)$tenant->datacenter->toArray();
    }

    /**
     * @param $login
     * @param $datacenterId
     * @return bool
     */
    public function loginExists($login, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        return $this->checkLoginExist($login, $datacenterId);
    }

    public function updateUsageAndLimits()
    {
        Log::getInstance()->event(
            static::LOG_METHOD_CALL,
            $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        );

        $this->updateCounterLimits();
        $this->_updateProfileLink();

        $tenantUsages = $this->fetchTenantLimitsAndUsages();

        return [
            static::TENANT => $tenantUsages,
            static::USER_USAGE => [],
            static::ERROR => null,
        ];
    }

    public function createAdministrator($parameters, $tenantId = '', $datacenterId = '')
    {
        if ($this->tenant->profile->isEmpty()) {
            $parameters->notifications = [
                UserNotification::BACKUP_DAILY_REPORT,
                UserNotification::QUOTA,
                UserNotification::REPORTS
            ];
        } else {
            $profile = $this->tenant->profile;
            $parameters->notifications = is_array($profile->notifications) ? $profile->notifications : [];
        }
        $tenantId = $tenantId ?: $this->tenant->tenantId;
        $adminInfo = parent::createAdministrator($parameters, $tenantId, $datacenterId);
        $mspApi = $this->tenant->getMspApi();
        try {
            $mspApi->postMailActivate(new Service_MspApiClient_Request_Struct_PostMailActivate(
                $adminInfo->getContact()->getEmail(), $adminInfo->getLogin()
            ));
        } catch (Exception $e) {
            Log::getInstance()->error($e->getMessage());
        }

        $this->throwNoContent();
    }

    public function getAdministratorConsole($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function adminUrlRedirect()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $mspApi = $this->getMspApi();
        $brand = $mspApi->getBrand($tenant->mspGroupId);
        $url = $brand->getAccountServerUrl();

        return (object)array(
            static::FIELD_URL => Str::format(static::REDIRECT_URL, $url, $tenant->mspGroupId)
        );
    }

    public function getDatacenters()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getServicePlans()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getServicePlan($planId)
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function setStorageForPlan($planId, $parameters)
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function unsetStorageForPlan($planId)
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getQuotasStatistics()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getUserGuideUrl()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $brand = $this->getMspApi()->getBrand($this->tenant->mspGroupId);
        return $brand->getUserGuideUrl();
    }

    /* ======= OTHER METHODS ======= */

    public function resolveDatacenterId()
    {
        $tenant = $this->tenant;
        // Do nothing if datacenter ID already was set from external side (online store / migration script)
        if (isset($tenant->datacenterId)) {
            return $tenant->datacenterId;
        }

        $datacenters = $tenant->global->datacenters;
        if ($datacenters->count() === 0) {
            throw new \Rest\RestException(401, static::EXCEPTION_NEED_TO_SET_CREDENTIALS_MESSAGE);
        }

        $profileDatacenters = $tenant->profile->datacenters;

        if (count($profileDatacenters) == 0) {
            throw new \Rest\RestException(400, Str::format(
                'Please specify datacenter in the service profile "%s".',
                $tenant->profile->resourceTypeId
            ));
        }
        if (count($profileDatacenters) == 1) {
            /** @var BackupProfileDatacenter $firstDatacenter */
            $firstDatacenter = reset($profileDatacenters);

            return $firstDatacenter->datacenterId;
        }

        $parametersManager = $this->getSubscriptionParametersManager();
        $datacenterId = $parametersManager->getParameterDatacenterId($tenant->subscriptionId);

        $datacenters = array_filter($profileDatacenters, function ($datacenter) use ($datacenterId) {
            return $datacenter->datacenterId == $datacenterId;
        });

        if (empty($datacenters)) {
            throw new \Rest\RestException(400, Str::format(
                'Please add the datacenter with ID "%s" to the list of datacenters in service profile "%s".',
                $datacenterId, $tenant->profile->resourceTypeId
            ));
        } else {
            /** @var BackupProfileDatacenter $firstDatacenter */
            $firstDatacenter = reset($datacenters);
            $datacenterId = $firstDatacenter->datacenterId;
        }

        return $datacenterId;
    }

    public function resolveBackupStorageId()
    {
        Log::getInstance()->event(
            static::LOG_METHOD_CALL,
            $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        );

        $this->throwNotApplicableMethodException();
    }

    /**
     * @return int
     */
    private function _createGroupForPartner()
    {
        $parentMspGroupId = $this->_getOrCreateParentGroupId();
        $tenant = $this->tenant;
        $mspGroupManager = $this->getMspGroupManager();
        $groupId = $mspGroupManager->createGroupForCustomer(
            $tenant, $parentMspGroupId, Service_MspApiClient_Request_Struct_CreateGroup::RESELLER_GROUP_KIND_ID
        );
        return $groupId;
    }

    protected function _updateGroupForPartner()
    {
        $this->getMspGroupManager()->updateGroupForCustomer($this->tenant);
    }

    /**
     * @throws Service_BaseClient_HTTPException
     */
    private function _deleteGroupForPartner()
    {

        $cloudApi = $this->createCloudApiInstance($this->tenant->datacenter);
        $tenantInfo = $cloudApi->getTenant($this->tenant->tenantId);


        if (is_null($tenantInfo)) {
            Log::getInstance()->warning(sprintf(self::LOG_WARNING_UNPROVISION_UNABLE_TO_FIND_GROUP, $this->tenant->tenantId));
            throw new \Exception(sprintf(self::LOG_WARNING_UNPROVISION_UNABLE_TO_FIND_GROUP, $this->tenant->tenantId));
        }

        $version = $tenantInfo->getVersion();

        if ($tenantInfo->getEnabled()) {
            $tenantInfo = $cloudApi->disableTenant($tenantInfo->getId(), $version);
            $version = $tenantInfo->getVersion();
        }

        $cloudApi->deleteTenant($tenantInfo->getId(), $version);
    }

    /**
     * @throws \Acronis\App\UsageReport\Exception\ManagerException
     * @throws \Rest\RestException
     */
    private function _retrieveUsageForPartner()
    {
        $this->updateCountersUsage($this->tenant->datacenter);
    }

    private function createAdministratorAccount()
    {
        $userInfo = $this->createAdministratorInfo();
        //add administrator account to baas with data from PA staff member
        $userManager = $this->getUserManager();
        $userManager->provisionAdministratorAccount(
            $this->tenant->mspGroupId,
            $this->tenant->subscriptionId,
            $userInfo
        );
    }

    /**
     * @return UserPost
     */
    private function createAdministratorInfo()
    {
        $apsApi = new Service_APSApi_Api($this->getApsc()->impersonate($this->tenant));

        $adminInfoFromPA = $apsApi->getFirstStaffMemberInfo($this->tenant);

        $admInfo = new UserPost();
        $admInfo->setTenantId($this->tenant->tenantId);
        $admInfo->setContact(new Contact());
        $admInfo->getContact()->setFirstName($adminInfoFromPA->givenName);
        $admInfo->getContact()->setLastName($adminInfoFromPA->familyName);
        $admInfo->getContact()->setEmail($adminInfoFromPA->email);

        $login = $this->createLoginForAdministrator($admInfo);
        $admInfo->setLogin($login);
        if ($this->tenant->profile->isEmpty()) {
            $admInfo->setNotifications([
                UserNotification::BACKUP_DAILY_REPORT,
                UserNotification::QUOTA,
                UserNotification::REPORTS
            ]);
        } else {
            $notifications = $this->tenant->profile->notifications;
            $admInfo->setNotifications(is_array($notifications) ? $notifications : []);
        }

        return $admInfo;
    }


    /**
     * @param string $string
     * @return string
     */
    private function clean($string)
    {
        return preg_replace('/[^A-Za-z0-9\-\_\@\.]/', '', $string); // Removes special chars.
    }

    /**
     * @param UserPost $adminInfo
     * @return string
     */
    protected function createLoginForAdministrator(UserPost $adminInfo)
    {
        $login = trim($adminInfo->getContact()->getFirstName() . '_' . $adminInfo->getContact()->getLastName(), '_');
        $login = preg_replace('/\s/', '_', $login);
        if (empty($login)) {
            $emails = explode('@', $adminInfo->getContact()->getEmail());
            $login = reset($emails);
        }
        $i = 1;
        $login = $this->clean($login);
        while (true) {
            $newLogin = $login . '_' . $i;
            if (!$this->loginExists($newLogin, $this->tenant->datacenterId)) {
                return $newLogin;
            }
            $i++;
        }
    }
}
