<?php
/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

class Service_PbaApi_Client implements Service_PbaApi_ClientInterface
{
    const HTTP_METHOD_POST = 'POST';
    const HTTP_URL_TEMPLATE = '%s://%s:%d/%s';
    const HTTP_URL_SCHEME_HTTPS = "https";
    const HTTP_URL_SCHEME_HTTP = "http";
    const HTTP_URL_DEFAULT_PORT = 5224;
    const HTTP_URL_DEFAULT_PATH = 'RPC2';

    const HTTP_CONTENT_TYPE_TEXT_XML = "Content-Type: text/xml";
    const HTTP_CODE_SUCCESS = 200;

    const XML_RPC_METHOD_EXECUTE = 'Execute';
    const XML_RPC_METHOD_COMMIT_TRANSACTION = 'CommitTransaction';
    const XML_RPC_METHOD_ROLLBACK_TRANSACTION = 'RollbackTransaction';

    const XML_RPC_PARAM_AUTO_COMMIT = 'AutoCommit';
    const XML_RPC_PARAM_USERNAME = 'Username';
    const XML_RPC_PARAM_PASSWORD = 'Password';
    const XML_RPC_PARAM_SERVER = 'Server';
    const XML_RPC_PARAM_METHOD = 'Method';
    const XML_RPC_PARAM_PARAMS = 'Params';
    const XML_RPC_PARAM_LANGUAGE = 'Lang';

    const XML_RPC_FIELD_FAULT_CODE = "faultCode";
    const XML_RPC_FIELD_FAULT_STRING = "faultString";
    const XML_RPC_FIELD_RESULT = 'Result';
    const XML_RPC_FIELD_TRANSACTION_ID = 'TransactionID';

    const SERVER_NAME_BM = 'BM';
    const AUTO_COMMIT_NO = 'No';

    const EXC_MSG_UNDEFINED_RESPONSE = 'Undefined response code (%s) with body: ​%s';
    const EXC_MSG_UNKNOWN_FAULT = 'Unknown fault';
    const EXC_MSG_PREVIOUS_TRANSACTION_IS_NOT_CLOSED = "The previous transaction is not closed.";

    /** @var string */
    private $_url;

    /** @var Service_PbaApi_HttpTransportInterface */
    private $_transport;

    /** @var Service_PbaApi_XmlRpcInterface */
    private $_xmlRpc;

    /** @var string */
    private $_serverName = self::SERVER_NAME_BM;

    /** @var string */
    private $_username;

    /** @var string */
    private $_password;

    /** @var bool */
    private $_autoCommit = true;

    /** @var integer */
    private $_transactionId;

    /**
     * @param string $host
     * @param int | null $port
     * @param string | null $path
     * @param boolean $secure
     * @param Service_PbaApi_HttpTransportInterface $httpTransport
     * @param Service_PbaApi_XmlRpcInterface $xmlRpc
     */
    public function __construct(
        $host, $port = NULL, $path = NULL, $secure = true,
        Service_PbaApi_HttpTransportInterface $httpTransport = NULL,
        Service_PbaApi_XmlRpcInterface $xmlRpc = NULL
    )
    {
        $this->_url = sprintf(
            self::HTTP_URL_TEMPLATE,
            $secure ? self::HTTP_URL_SCHEME_HTTPS : self::HTTP_URL_SCHEME_HTTP,
            $host,
            empty($port) ? self::HTTP_URL_DEFAULT_PORT : $port,
            empty($path) ? self::HTTP_URL_DEFAULT_PATH : $path
        );
        $this->_transport = is_null($httpTransport) ? new Service_PbaApi_HttpTransport() : $httpTransport;
        $this->_xmlRpc = is_null($xmlRpc) ? new Service_PbaApi_XmlRpc() : $xmlRpc;
    }

    /**
     * @param string $serverName
     */
    public function setServerName($serverName)
    {
        $this->_serverName = (string)$serverName;
    }

    /**
     * @param string $username
     * @param string $password
     */
    public function setCredentials($username, $password)
    {
        $this->_username = (string)$username;
        $this->_password = (string)$password;
    }

    public function startTransaction()
    {
        Log::getInstance()->event("Start Transaction");
        if (!is_null($this->_transactionId)) {
            throw new Service_PbaApi_Client_TransactionException(self::EXC_MSG_PREVIOUS_TRANSACTION_IS_NOT_CLOSED);
        }
        $this->_autoCommit = false;
    }

    public function commitTransaction()
    {
        Log::getInstance()->event("Commit Transaction");
        if (!is_null($this->_transactionId)) {
            $arguments = array(
                self::XML_RPC_FIELD_TRANSACTION_ID => $this->_transactionId
            );
            $result = $this->_sendRequest(self::XML_RPC_METHOD_COMMIT_TRANSACTION, $arguments);
            Log::getInstance()->event($result);
        }
        $this->_transactionId = null;
        $this->_autoCommit = true;
    }

    public function rollbackTransaction()
    {
        Log::getInstance()->event("Rollback Transaction");
        if (!is_null($this->_transactionId)) {
            $arguments = array(
                self::XML_RPC_FIELD_TRANSACTION_ID => $this->_transactionId
            );
            $result = $this->_sendRequest(self::XML_RPC_METHOD_ROLLBACK_TRANSACTION, $arguments);
            Log::getInstance()->event($result);
        }
        $this->_transactionId = null;
        $this->_autoCommit = true;
    }

    /**
     * @param Service_PbaApi_Client_RequestInterface $request
     * @return Service_PbaApi_Client_ResponseInterface
     *
     * @throws Exception
     * @throws Service_PbaApi_Client_HttpException
     * @throws Service_PbaApi_Client_XmlRpcException
     */
    public function execute(Service_PbaApi_Client_RequestInterface $request)
    {
        $arguments = array();
        if (!$this->_autoCommit) {
            $arguments[self::XML_RPC_PARAM_AUTO_COMMIT] = self::AUTO_COMMIT_NO;
            if (!is_null($this->_transactionId)) {
                $arguments[self::XML_RPC_FIELD_TRANSACTION_ID] = $this->_transactionId;
            }
        }

        if (!is_null($this->_username)) {
            $arguments[self::XML_RPC_PARAM_USERNAME] = (string)$this->_username;
            $arguments[self::XML_RPC_PARAM_PASSWORD] = (string)$this->_password;
        }

        if (!is_null($this->_serverName)) {
            $arguments[self::XML_RPC_PARAM_SERVER] = (string)$this->_serverName;
        }

        $arguments[self::XML_RPC_PARAM_METHOD] = $request->getMethod();
        $arguments[self::XML_RPC_PARAM_PARAMS] = $request->getParameters();
        if (!is_null($request->getLanguage())) {
            $arguments[self::XML_RPC_PARAM_LANGUAGE] = $request->getLanguage();
        }

        $responseArray = $this->_sendRequest(self::XML_RPC_METHOD_EXECUTE, $arguments);

        if (!$this->_autoCommit && is_null($this->_transactionId)) {
            $this->_transactionId = $responseArray[self::XML_RPC_FIELD_TRANSACTION_ID];
        }

        $result = isset($responseArray[self::XML_RPC_FIELD_RESULT]) ? $responseArray[self::XML_RPC_FIELD_RESULT] : NULL;
        return $request->getExpectedResponse($result);
    }

    /**
     * @param string $method
     * @param array $arguments
     * @return array|NULL
     * @throws Service_PbaApi_Client_HttpException
     * @throws Service_PbaApi_Client_XmlRpcException
     */
    private function _sendRequest($method, $arguments)
    {
        $requestBody = $this->_xmlRpc->encodeRequest($method, $arguments, array(
            'encoding' => 'UTF-8',
            'escaping' => 'markup'
        ));

        $response = $this->_transport->sendRequest(new Service_PbaApi_HttpTransport_Request(
            self::HTTP_METHOD_POST,
            $this->_url,
            $requestBody,
            array(self::HTTP_CONTENT_TYPE_TEXT_XML)
        ));

        $httpCode = $response->getHttpCode();
        $responseBody = $response->getBody();
        if ($httpCode !== self::HTTP_CODE_SUCCESS) {
            $message = sprintf(self::EXC_MSG_UNDEFINED_RESPONSE, $httpCode, $responseBody);
            Log::getInstance()->error($message);
            throw new Service_PbaApi_Client_HttpException($message, $httpCode);
        }
        $responseArray = $this->_xmlRpc->decode($responseBody);
        if (is_array($responseArray) && $this->_xmlRpc->isFault($responseArray)) {
            $faultCode = isset($responseArray[self::XML_RPC_FIELD_FAULT_CODE]) ?
                $responseArray[self::XML_RPC_FIELD_FAULT_CODE] : 0;
            $faultString = isset($responseArray[self::XML_RPC_FIELD_FAULT_STRING]) ?
                base64_decode($responseArray[self::XML_RPC_FIELD_FAULT_STRING]) : self::EXC_MSG_UNKNOWN_FAULT;

            Log::getInstance()->event($faultCode . '-' . $faultString);
            throw new Service_PbaApi_Client_XmlRpcException($faultString, $faultCode);
        }
        return $responseArray;
    }
}