<?php
/**
 * @Copyright © 2002-2020 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\Services\CountersManager;

use Acronis\App\Base\ApplicationCounter;
use Acronis\App\Base\ApplicationCounter as Counter;
use Acronis\App\CloudApi\ApiInterface;
use Acronis\App\Services\CountersManager\Filters\ExcludeNonUserCounters;
use Acronis\App\Services\CountersManager\Filters\ExcludeStorageAdditionalCounters;
use Acronis\App\Services\CountersManager\Filters\StdEditionCountersOnly;
use Acronis\App\Utils\MemoizeTrait;
use Acronis\App\Utils\UomConverter as Uom;
use Arr;
use Symfony\Component\Yaml\Yaml;

class CountersManager
{
    use MemoizeTrait;

    const FILTER_EXCLUDE_STORAGE_ADDITIONAL_COUNTERS = 1;
    const FILTER_STD_EDITION_COUNTERS_ONLY = 2;
    const FILTER_EXCLUDE_NON_USER_COUNTERS = 4;

    /**
     * This is a map of filter bit mask to class implementing CounterFilterInterface
     * @var string[]
     */
    private static $availableFilters = [
        self::FILTER_EXCLUDE_STORAGE_ADDITIONAL_COUNTERS => ExcludeStorageAdditionalCounters::class,
        self::FILTER_STD_EDITION_COUNTERS_ONLY => StdEditionCountersOnly::class,
        self::FILTER_EXCLUDE_NON_USER_COUNTERS => ExcludeNonUserCounters::class,
    ];

    /** @var array */
    private $applications;

    /** @var Counter[] */
    private $countersHash;

    /** @var Counter[] */
    private $offeringItemsHash;

    /**
     * Counters constructor.
     *
     * @param array $countersData
     */
    public function __construct(array $countersData)
    {
        $this->countersHash = [];
        $this->offeringItemsHash = [];
        $this->applications = [];

        foreach ($countersData as $counterData) {
            $counter = new Counter($counterData);

            $this->countersHash[$counter->getName()] = $counter;
            $this->offeringItemsHash[$counter->getOfferingItemName()] = $counter;
            foreach ($counter->getPartnerOfferingItems() as $partnerOfferingItem) {
                $partnerCounter = clone $counter;
                $partnerCounter->setTypesPartnerAndReseller();
                $this->offeringItemsHash[$partnerOfferingItem] = $partnerCounter;
            }
            $this->applications[$counter->getApplicationType()][] = $counter;
        }
    }

    public function getUserApplicableOfferingItemsSet()
    {
        return array_reduce($this->getAllCounters(), function ($set, $counter) {
            /** @var ApplicationCounter $counter */
            if ($counter->isForUser()) {
                $set[$counter->getOfferingItemName()] = true;
            }

            return $set;
        }, []);
    }

    /**
     * @param string $counterName
     * @return Counter|null
     */
    public function getCounterByName($counterName)
    {
        return Arr::get($this->countersHash, $counterName, null);
    }

    /**
     * @param string $offeringItemName
     * @return Counter|null
     */
    public function getCounterByOfferingItemName($offeringItemName)
    {
        return Arr::get($this->offeringItemsHash, $offeringItemName, null);
    }

    public function hasUsageTypeDuplicationInBoundsOfEdition($offeringItemName)
    {
        $counter = $this->getCounterByOfferingItemName($offeringItemName);
        $edition = $counter->getEdition();
        $usageType = $counter->getUsageType();

        $countersFromSameEdition = $this->memoize(function() use ($edition) {
            return $this->getAllCounters(static::FILTER_EXCLUDE_STORAGE_ADDITIONAL_COUNTERS, $edition);
        }, $edition);

        $equalUsageTypes = array_filter($countersFromSameEdition, function($counter) use ($usageType) {
            /** @var Counter $counter */
            return $counter->getUsageType() === $usageType;
        });

        return count($equalUsageTypes) > 1;
    }

    /**
     * @param string[] $offeringItemsNames
     * @param callable $skipCondition
     * @return Counter[]
     */
    public function findCountersByOfferingItemNames($offeringItemsNames, $skipCondition = null)
    {
        return array_values(array_filter($this->countersHash,
            function ($counter) use ($offeringItemsNames, $skipCondition) {
                /** @var  ApplicationCounter $counter */
                return (!$skipCondition || $skipCondition($counter)) && in_array($counter->getOfferingItemName(),
                        $offeringItemsNames);
            }));
    }

    /**
     * @param string $applicationType
     * @return Counter[]
     */
    public function getApplicationCounters($applicationType)
    {
        return Arr::get($this->applications, $applicationType, []);
    }

    /**
     * Supports filtering of counters list
     * @param int $filterFlags
     * @param null $edition
     * @return Counter[]
     */
    public function getAllCounters($filterFlags = 0, $edition = null)
    {
        $counters = array_values($this->countersHash);
        foreach (static::$availableFilters as $filterBitMask => $filterClassName) {
            if ($filterFlags & $filterBitMask) {
                /** @var CounterFilterInterface $filter */
                $filter = new $filterClassName();
                $counters = array_filter($counters, function($counter) use ($filter, $edition) {
                   /** @var Counter $counter */
                   // Optional filtering by counter edition
                   if ($edition !== null && $counter->getEdition() !== $edition) {
                       return false;
                   }
                   return !$filter->needsToBeExcluded($counter);
                });
            }
        }

        return $counters;
    }

    /**
     * @return array
     */
    public function getAllCountersMap()
    {
        $applicationCounters = Yaml::parseFile(DIR_APP . DS . 'application_counters.yaml');
        $stdCounters = array_filter($applicationCounters, function ($counter) {
            return Arr::get($counter, 'edition', ApiInterface::EDITION_STD) === ApiInterface::EDITION_STD
                && Arr::get($counter, 'counter_type') !== 'usage';
            // last check if for skipping counters for usage only like AcronisHostedStorage
        });
        $usageNamesToStdCounters = Arr::map($stdCounters, 'usage_type', 'counter_name');
        $allCounters = [];

        foreach ($applicationCounters as $counter) {
            $originalCounterName = $counter['counter_name'];
            $stdCounterName = Arr::get($usageNamesToStdCounters, $counter['usage_type']);
            $counterName = $stdCounterName ?: $originalCounterName;
            $allCounters[$counterName] = Arr::get($allCounters, $counterName, []);
            $allCounters[$counterName][] = [
                'name' => $originalCounterName,
                'is_for_user' => Arr::get($counter, 'is_for_user', false)
            ];
        }

        return [
            'counters' => $allCounters,
            'counters_meta' => array_map([$this, 'buildCounterMetaForUI'], $this->countersHash)
        ];
    }

    /**
     * @return array
     */
    public function buildOfferingItemNameNormalizationMap()
    {
        $normalizationMap = [];
        foreach ($this->getAllCounters() as $counter) {
            $usageType = $counter->getUsageType();
            $offeringItemName = $counter->getOfferingItemName();
            $normalizationMap[$offeringItemName] = str_replace(['pack_adv_', 'dr_'], '', $usageType);
        }

        return $normalizationMap;
    }

    public function convertValueToOfferingItemUnits(Counter $counter, $counterValue)
    {
        if ($counterValue === 0) {
            return 0;
        }

        if ($counterValue === Counter::UNLIMITED_VALUE) {
            return ApiInterface::UNLIMITED_OFFERING_ITEM_VALUE;
        }

        return Uom::convert($counterValue, $counter->getUnit(), $counter->getOfferingItemUnit());
    }

    public function convertValueFromOfferingItemUnits(Counter $counter, $offeringItemValue)
    {
        if ($offeringItemValue === 0) {
            return 0;
        }

        if ($offeringItemValue === ApiInterface::UNLIMITED_OFFERING_ITEM_VALUE) {
            return Counter::UNLIMITED_VALUE;
        }

        return Uom::convert($offeringItemValue, $counter->getOfferingItemUnit(), $counter->getUnit());
    }

    /**
     * @param Counter $counter
     * @return array
     */
    private function buildCounterMetaForUI($counter)
    {
        return [
            'counter_name' => $counter->getName(),
            'offering_item_name' => $counter->getOfferingItemName(),
            'unit_of_measure' => $counter->getUnit(),
            'edition' => $counter->getEdition(),
            'sorting_index' => $counter->getSortingIndex(),
            'is_feature' => $counter->isFeature()
        ];
    }
}