<?php

use Acronis\App\Services\Config\ConfigFactory;
use Acronis\App\Services\Config\ConfigSchema;
use Acronis\App\Services\Locator;

/**
 * @Copyright © 2002-2020 Acronis International GmbH. All rights reserved
 */

class Errors
{
    const ERROR_MESSAGE_FORMAT = '%s (%d): %s in file %s on line %d';
    const FATAL_ERROR_MESSAGE_FORMAT = 'PHP FATAL (%s):  %s in %s on line %d';

    private $_errNo = 0;
    private $_errStr = '';
    private $_errFile = '';
    private $_errLine;
    private $_errContext = array();
    private $_severity = '';

    static public function setHandlers()
    {
        set_error_handler(array(new Errors, 'handler'));
        register_shutdown_function(array(new Errors, 'fatalHandler'));
    }

    /**
     * @param int $errNo
     * @param string $errStr
     * @param string $errFile
     * @param int $errLine
     * @param array $errContext
     * @throws Exception
     * @return bool
     */
    public function handler($errNo, $errStr, $errFile, $errLine, array $errContext = array())
    {
        if (!(error_reporting() & $errNo))
        {
            return true;
        }

        $this->_errNo = $errNo;
        $this->_errStr = $errStr;
        $this->_errFile = $errFile;
        $this->_errLine = $errLine;
        $this->_errContext = $errContext;

        $this->_loadSeverity();

        $message = sprintf(
            self::ERROR_MESSAGE_FORMAT,
            $this->_severity,
            $this->_errNo,
            $this->_errStr,
            $this->_errFile,
            $this->_errLine
        );

        Log::getInstance()->error($message);
        throw new Exception($message, $this->_errNo);
    }

    /**
     * Fatal error handler. Registered via register_shutdown_function()
     * @return void
     */
    public function fatalHandler()
    {
        $error = error_get_last();
        $type = Arr::get($error, 'type', '');
        if (!is_numeric($type) || $type == E_DEPRECATED) {
            return;
        }
        Log::getInstance()->error(
            self::FATAL_ERROR_MESSAGE_FORMAT,
            Arr::get($error, 'type'),
            Arr::get($error, 'message', 'Unknown error'),
            Arr::get($error, 'file'),
            Arr::get($error, 'line')
        );
    }

    /**
     * Set error_type for specified error code
     * @return void
     */
    private function _loadSeverity()
    {
        switch ($this->_errNo)
        {
            case E_NOTICE :
            case E_USER_NOTICE :
                $this->_severity = 'NOTICE';
                break;
            case E_WARNING :
            case E_USER_WARNING :
            case E_CORE_WARNING :
            case E_COMPILE_WARNING :
                $this->_severity = 'WARNING';
                break;
            case E_ERROR :
            case E_USER_ERROR :
            case E_RECOVERABLE_ERROR :
                $this->_severity = 'FATAL ERROR';
                break;
            case E_PARSE :
                $this->_severity = 'PARSE ERROR';
                break;
            case E_CORE_ERROR :
            case E_COMPILE_ERROR :
                $this->_severity = 'INTERNAL ERROR';
                break;
            case E_STRICT :
                $this->_severity = 'PHP STRICT';
                break;
            case E_DEPRECATED :
            case E_USER_DEPRECATED :
                $this->_severity = 'PHP DEPRECATED';
                break;
            default:
                $this->_severity = 'UNKNOWN';
                break;
        }
        return;
    }
}
