<?php
/**
 * @Copyright © 2002-2020 Acronis International GmbH. All rights reserved
 */

use Acronis\App\Services\Apsc\ProviderApscAwareTrait;
use Rest\RestException;

require_once(__DIR__ . '/app/bootstrap.php');

/**
 * Class globals
 * @type("http://acronis.com/aps/cloud-backup-v1/globals/9.6")
 * @implements("http://aps-standard.org/types/core/application/1.0")
 * @access("global", true)
 */
class globals extends APSResource
{
    use ProviderApscAwareTrait;

    const PACKAGES_URI = '/aps/2/packages';
    const ERROR_ENDPOINT_OUTDATED = 'Outdated endpoint version. Please update endpoint first.';

    const LOG_CALLED_PROVISION = 'Application provision was called.';
    const LOG_CALLED_UPGRADE = 'Application upgrade from version %s was called.';
    const LOG_CALLED_GET_PARTNER_CREDENTIALS = 'Get partner credentials was called.';
    const LOG_CALLED_ON_TENANT_AVAILABLE = 'onTenantAvailable was called.';
    const LOG_ERROR_JSON_DECODE = 'Can\'t decode JSON. Error: %s %s; Input: "%s"';

    const ON_TENANT_AVAILABLE = "onTenantAvailable";

    const EX_MSG_JSON_DECODE_ERROR = 'Invalid JSON parse input';
    const EX_MSG_WRONG_PA_AND_RUNTIME_VERSION = 'Please upgrade PA to version 6.0 or later. Newer version of APS PHP Runtime is required. The latest version of the APS PHP Runtime is available for download at the following URL: http://doc.apsstandard.org/2.1/tools/php-lib/#php-runtime-download.';
    const EX_MSG_ASSIGN_DISCOUNT_FOR_L1_RESELLER_INVALID_DISCOUNT_ID = 'Property "discountId" is not specified or invalid. It is required and should be a positive integer number.';
    const EX_MSG_ASSIGN_DISCOUNT_FOR_L1_RESELLER_INVALID_ACCOUNT_ID = 'Property "accountId" is invalid. It should be a positive integer number.';
    const EX_MSG_L1_RESELLER_IS_NOT_FOUND = 'First-level reseller is not found.';
    const EX_MSG_DISCOUNT_IS_NOT_FOUND = 'Discount is not found.';
    const EX_MSG_DISCOUNT_IS_NOT_ALLOWED_TO_USE_FOR_L1_RESELLER = 'Discount "%s" is not allowed to use.';
    const EX_MSG_DISCOUNT_DOES_NOT_BELONG_TO_RESELLER = 'Discount "%s" does not belong to reseller "%s".';
    const EX_MSG_MINIMUM_COMMITMENT_IS_OFF = 'Minimum commitment is off.';

    const STRING_EMPTY = '';
    const JSON_UNACCEPTED_CHARS = '(?! )\s';
    const JSON_LAST_ERROR_MSG_FUNCTION = 'json_last_error_msg';

    const RUNTIME_2_1 = '2.1';
    const DEFAULT_DATACENTER_ID = '1';

    /* ======= RESOURCE PROPERTIES ======= */

    /**
     * @link("http://acronis.com/aps/cloud-backup-v1/tenant/9.0[]")
     */
    public $tenant;

    /**
     * @link("http://acronis.com/aps/cloud-backup-v1/backup-profile/5.0[]")
     */
    public $storages; //profiles

    /**
     * @link("http://acronis.com/aps/cloud-backup-v1/datacenter/4.0[]")
     */
    public $datacenters;

    /**
     * @link("http://acronis.com/aps/cloud-backup-v1/minimal-commitment-tier/4.0[]")
     */
    public $minimalCommitmentTiers;

    /**
     * @type("boolean")
     */
    public $minimalCommitment = false;

    /**
     * @type("boolean")
     * @readonly
     */
    public $isInitialised = false;

    /**
     * @type(ResellerRelation[])
     */
    public $resellerRelations = array();

    /**
     * @type(DiscountRelation[])
     */
    public $discountRelations = array();

    /**
     * @type(AccountToProcess[])
     */
    public $accountToProcess = array();

    /* ======= OTHER PROPERTIES ======= */

    /* ======= RESOURCE METHODS ======= */

    public function provision()
    {
        Log::getInstance()->event(self::LOG_CALLED_PROVISION);
    }

    /**
     * @verb(GET)
     * @path("/allCounters")
     * @return(array)
     */
    public function getAllCountersMap()
    {
        return $this->getAllCounters();
    }

    /**
     *
     * @param null $version
     * @throws Exception
     * @return void
     */
    public function upgrade($version = null)
    {
        Log::getInstance()->event(sprintf(self::LOG_CALLED_UPGRADE, $version));
        $buildFrom = substr($version, strpos($version, '-') + 1);
        ini_set('max_execution_time', 0);

        if ($this->_checkEndpointVersion()) {
            throw new Exception(self::ERROR_ENDPOINT_OUTDATED);
        }
        if ($this->getApsRuntimePackageVersion() < self::RUNTIME_2_1) {
            throw new RestException(409, self::EX_MSG_WRONG_PA_AND_RUNTIME_VERSION);
        }
        $this->_checkResourcesStatus();

        $this->aps->type = $this->getSelfApsType();
        if (is_null($this->minimalCommitment)) {
            $this->minimalCommitment = false;
        }
        if (is_array($this->resellerRelations)) {
            /**
             * @var ResellerRelation $relation
             */
            foreach ($this->resellerRelations as $relation) {
                if (!isset($relation->datacenterId)) {
                    $relation->datacenterId = self::DEFAULT_DATACENTER_ID;
                }
            }
        }

        unset($this->isInitialised);

        // remove deprecated properties
        $partnerCredentials = null;
        if (isset($this->partnerCredentials)) {
            $partnerCredentials = $this->partnerCredentials;
            unset($this->partnerCredentials);
        }
        if (isset($this->baseClientLink)) {
            unset($this->baseClientLink);
        }
        if (isset($this->clientLinks)) {
            unset($this->clientLinks);
        }
        if (isset($this->baseWinClientLink)) {
            unset($this->baseWinClientLink);
        }
        if (isset($this->baseLinClientLink)) {
            unset($this->baseLinClientLink);
        }
        if (isset($this->baseMacClientLink)) {
            unset($this->baseMacClientLink);
        }
        if (isset($this->branding)) {
            unset($this->branding);
        }
        $this->update();

        $this->_upgradeDatacenters();
        $this->_createDatacenterFromPartnerCredentials($partnerCredentials);
        $this->_upgradeProfiles();
        $this->_upgradeMinimalCommitmentTiers();
        if (intval($buildFrom) < 2228) {
            $this->_upgradeTenants();
        }

        $this->_unsubscribeFromOnTenantAvailableEvent();

        if (!count($this->storages)) {
            $this->_getApsApi()->makeDefaultProfile($this);
        }
        $this->fillResource($this->aps->id);
    }

    /* ======= RESOURCE EVENTS ======= */

    /* ======= RESOURCE OPERATIONS ======= */

    /**
     * @verb(GET)
     * @path("/firstLevelResellers")
     * @return(array)
     */
    public function getFirstLevelResellers()
    {
        Log::getInstance()->event(self::LOG_CALLED_METHOD_WAS_CALLED, __METHOD__);

        $this->checkDiscountManagementAccess();

        $relationsDict = array();
        foreach ($this->_getDiscountRelations() as $relation) {
            $relationsDict[$relation->accountId] = $relation->discountId;
        }

        $mcDiscountManager = $this->_managerFactory->getMinimalCommitmentDiscountManager(
            $this->_getOpenApi(),
            $this->_getPbaApi()
        );

        $result = array();
        foreach ($mcDiscountManager->fetchFirstLevelResellersAccounts() as $account) {
            $discountId = isset($relationsDict[$account->getAccountId()])
                ? $relationsDict[$account->getAccountId()]
                : null;
            $discountForResult = null;
            if ($discountId) {
                $discount = $mcDiscountManager->fetchDiscountById($discountId);
                if ($discount) {
                    $discountForResult = $this->_prepareDiscountForClient($discount);
                }
            }
            $result [] = array(
                'accountId' => $account->getAccountId(),
                'name' => $account->getCompanyName(),
                'discountId' => $discountId,
                'discount' => $discountForResult,
            );
        }
        return $result;
    }

    /**
     * @verb(GET)
     * @path("/discount/{discountId}")
     * @param(integer,path)
     * @return(object)
     */
    public function getDiscount($discountId)
    {
        Log::getInstance()->event(self::LOG_CALLED_METHOD_WAS_CALLED, __METHOD__);

        $this->checkDiscountManagementAccess();

        $mcDiscountManager = $this->_managerFactory->getMinimalCommitmentDiscountManager(
            $this->_getOpenApi(),
            $this->_getPbaApi()
        );
        $discount = $mcDiscountManager->fetchDiscountById($discountId);
        if (!$discount) {
            throw new RestException(404, self::EX_MSG_DISCOUNT_IS_NOT_FOUND);
        }
        return $this->_prepareDiscountForClient($discount);
    }

    /**
     * @verb(PUT)
     * @path("/reseller/{accountId}/discount")
     * @param(integer,path)
     * @param(object,body)
     * @return(object)
     */
    public function setDiscountForFirstLevelReseller($accountId, $inputData)
    {
        Log::getInstance()->event(self::LOG_CALLED_METHOD_WAS_CALLED, __METHOD__);

        $this->checkDiscountManagementAccess();

        if (!$this->minimalCommitment) {
            throw new RestException(409, self::EX_MSG_MINIMUM_COMMITMENT_IS_OFF);
        }

        if (filter_var($accountId, FILTER_VALIDATE_INT, array('min_range' => 0)) === false) {
            throw new RestException(400, self::EX_MSG_ASSIGN_DISCOUNT_FOR_L1_RESELLER_INVALID_ACCOUNT_ID);
        }

        if (
            !isset($inputData->discountId)
            || filter_var($inputData->discountId, FILTER_VALIDATE_INT, array('min_range' => 0)) === false
        ) {
            throw new RestException(400, self::EX_MSG_ASSIGN_DISCOUNT_FOR_L1_RESELLER_INVALID_DISCOUNT_ID);
        }

        $mcDiscountManager = $this->_managerFactory->getMinimalCommitmentDiscountManager(
            $this->_getOpenApi(),
            $this->_getPbaApi()
        );

        $account = $mcDiscountManager->fetchFirstLevelResellerAccountById((int)$accountId);
        if (!$account) {
            throw new RestException(400, self::EX_MSG_L1_RESELLER_IS_NOT_FOUND);
        }

        $discount = $mcDiscountManager->fetchDiscountById((int)$inputData->discountId);
        if (!$discount) {
            throw new RestException(400, self::EX_MSG_DISCOUNT_IS_NOT_FOUND);
        }
        if (!$mcDiscountManager->detectAllowingDiscountForFirstLevelReseller($discount)) {
            throw new RestException(
                400,
                Str::format(self::EX_MSG_DISCOUNT_IS_NOT_ALLOWED_TO_USE_FOR_L1_RESELLER, $discount->getId())
            );
        }

        if ($discount->getOwnerAccountId() != $account->getAccountId()) {
            throw new RestException(
                400,
                Str::format(
                    self::EX_MSG_DISCOUNT_DOES_NOT_BELONG_TO_RESELLER,
                    $discount->getId(),
                    $account->getAccountId()
                )
            );
        }

        $index = $this->_searchDiscountRelationByAccountId($account->getAccountId());
        if ($index === false) {
            $newRelation = new DiscountRelation();
            $newRelation->accountId = $account->getAccountId();
            $newRelation->discountId = $discount->getId();
            $this->discountRelations = $this->discountRelations ?: array();
            $this->discountRelations [] = $newRelation;
        } else {
            $this->discountRelations[$index]->discountId = $discount->getId();
        }
        $this->update();

        throw new RestException(200);
    }

    /**
     * @verb(DELETE)
     * @path("/reseller/{accountId}/discount")
     * @param(integer,path)
     * @return(object)
     */
    public function unsetDiscountForFirstLevelReseller($accountId)
    {
        Log::getInstance()->event(self::LOG_CALLED_METHOD_WAS_CALLED, __METHOD__);

        $this->checkDiscountManagementAccess();

        if (!$this->minimalCommitment) {
            throw new RestException(409, self::EX_MSG_MINIMUM_COMMITMENT_IS_OFF);
        }

        if (filter_var($accountId, FILTER_VALIDATE_INT, array('min_range' => 0)) === false) {
            throw new RestException(400, self::EX_MSG_ASSIGN_DISCOUNT_FOR_L1_RESELLER_INVALID_ACCOUNT_ID);
        }

        $mcDiscountManager = $this->_managerFactory->getMinimalCommitmentDiscountManager(
            $this->_getOpenApi(),
            $this->_getPbaApi()
        );

        $account = $mcDiscountManager->fetchFirstLevelResellerAccountById((int)$accountId);
        if (!$account) {
            throw new RestException(400, self::EX_MSG_L1_RESELLER_IS_NOT_FOUND);
        }

        $index = $this->_searchDiscountRelationByAccountId($account->getAccountId());
        if ($index !== false) {
            array_splice($this->discountRelations, $index, 1);
            $this->update();
        }

        throw new RestException(200);
    }

    /**
     * @verb(POST)
     * @path("/check_connection")
     * @param(object, body)
     */
    public function checkConnection($credentials)
    {
        $publicApiHost = defined('OPEN_API_URI') ? OPEN_API_URI : parse_url(\APS\Proto::getControllerURI(), PHP_URL_HOST);
        $connectionChecker = new Model_ConnectionChecker(
            $credentials->login,
            $credentials->password,
            $credentials->url,
            $publicApiHost
        );

        $checker = $connectionChecker->getCheckerByType(Model_ConnectionChecker::VERIFY_BAAS_CREDENTIALS);
        $success = $checker->check();

        if (!$success) {
            $errorInfo = $checker->getCheckError();
            throw new RestException($errorInfo->code, $errorInfo->message);
        }
    }

    /**
     * @verb(GET)
     * @path("/settings")
     */
    public function getSettings()
    {
        $config = Settings::getInstance()->getApplicationConfig();
        $settings = new stdClass();
        $settings->ccpv1_show_staff_members = $config->ccpv1_show_staff_members;
        return $settings;
    }

    /* ======= OTHER METHODS ======= */

    /**
     * @return string
     */
    protected function getSelfApsType()
    {
        return self::CLOUD_BACKUP_GLOBALS_TYPE;
    }

    /**
     * @return tenant
     */
    protected function createTenant()
    {
        return new tenant(
            $this->_facade,
            $this->_managerFactory,
            $this->_sessionStore,
            $this->_apsc
        );
    }

    /**
     * @return profile
     */
    protected function createProfile()
    {
        return new profile(
            $this->_facade,
            $this->_managerFactory,
            $this->_sessionStore,
            $this->_apsc
        );
    }

    private function _unsubscribeFromOnTenantAvailableEvent()
    {
        $this->unsubscribeFromEventByType(\APS\EventSubscription::Available, self::CLOUD_BACKUP_TENANT_TYPE);
    }

    /**
     * @param int $accountId
     * @return bool|int Index of found element
     */
    private function _searchDiscountRelationByAccountId($accountId)
    {
        foreach ($this->_getDiscountRelations() as $i => $relation) {
            if ($relation->accountId == $accountId) {
                return $i;
            }
        }
        return false;
    }

    /**
     * @return array
     */
    private function _getDiscountRelations()
    {
        return $this->discountRelations ?: array();
    }

    /**
     * @param Service_PbaApi_Api_Response_Struct_Discount $discount
     * @return array
     */
    private function _prepareDiscountForClient(Service_PbaApi_Api_Response_Struct_Discount $discount)
    {
        $mcDiscountManager = $this->_managerFactory->getMinimalCommitmentDiscountManager(
            $this->_getOpenApi(),
            $this->_getPbaApi()
        );
        return array(
            'id' => $discount->getId(),
            'name' => $discount->getName(),
            'ownerAccountId' => $discount->getOwnerAccountId(),
            'type' => $discount->getType(),
            'status' => $discount->getStatus(),
            'rate' => $discount->getRate(),
            'isForAllPlans' => $discount->getIsForAllPlans(),
            'isForAllResources' => $discount->getIsForAllResources(),
            'isAllowedForFirstLevelReseller' => $mcDiscountManager->detectAllowingDiscountForFirstLevelReseller($discount),
        );
    }

    /**
     * @param string $tenantId
     */
    protected function upgradeTenant($tenantId)
    {
        $tenant = $this->createTenant();
        $tenant->fillResource($tenantId);
        $tenant->upgrade();
    }

    /**
     * @param string $profileId
     */
    protected function upgradeMinimalCommitmentTier($profileId)
    {
        $tier = new minimalCommitmentTier(
            $this->_facade,
            $this->_managerFactory,
            $this->_sessionStore,
            $this->_apsc
        );
        $tier->fillResource($profileId);
        $tier->upgrade();
    }

    /**
     * @param string $profileId
     */
    protected function upgradeProfile($profileId)
    {
        $profile = $this->createProfile();
        $profile->fillResource($profileId);
        $profile->upgrade();
    }

    /**
     * @param string $datacenterId
     */
    protected function upgradeDatacenter($datacenterId)
    {
        $profile = new datacenter(
            $this->_facade,
            $this->_managerFactory,
            $this->_sessionStore,
            $this->_apsc
        );
        $profile->fillResource($datacenterId);
        $profile->upgrade();
    }

    private function _upgradeTenants()
    {
        $tenants = $this->_getApsApi()->getTenantsForGlobal($this);
        foreach ($tenants as $tenant) {
            $this->upgradeTenant($tenant->aps->id);
        }
    }

    private function _upgradeProfiles()
    {
        $profiles = $this->_getProfiles();
        foreach ($profiles as $profile) {
            $this->upgradeProfile($profile->aps->id);
        }
    }

    private function _getProfiles()
    {
        $profiles = $this->storages;
        if (!count($profiles)) {
            $profiles = $this->_getApsApi()->getProfilesForGlobal($this);
        }
        return $profiles;
    }

    private function _upgradeMinimalCommitmentTiers()
    {
        $tiers = $this->_getMinimalCommitmentTiers();
        foreach ($tiers as $tier) {
            $this->upgradeMinimalCommitmentTier($tier->aps->id);
        }
    }

    private function _getMinimalCommitmentTiers()
    {
        $tiers = $this->minimalCommitmentTiers;
        if (!count($tiers)) {
            $tiers = $this->_getApsApi()->getMinimalCommitmentTiersForGlobal($this);
        }
        return $tiers;
    }

    private function _upgradeDatacenters()
    {
        $datacenters = $this->_getDatacenters();
        foreach ($datacenters as $datacenter) {
            $this->upgradeDatacenter($datacenter->aps->id);
        }
    }

    private function _getDatacenters()
    {
        $datacenters = $this->datacenters;
        if (!count($datacenters)) {
            $datacenters = $this->_getApsApi()->getDatacentersForGlobal($this);
        }
        return $datacenters;
    }

    /**
     * @return bool
     */
    protected function _checkEndpointVersion()
    {
        $endpointVersion = $this->_getBuildInfo()->getPackageBuildNumber();

        //ToDo: move to Service_APSApi_Api
        $packages = $this->_safe_json_decode($this->_apsc->getIo()->sendRequest(\APS\Proto::GET, self::PACKAGES_URI));
        $versions = array();
        foreach ($packages as $package) {
            if ($package->id !== self::PACKAGE_ID) {
                continue;
            }
            $versions[] = intval($package->release);
        }
        $packageVersion = max($versions);

        return $endpointVersion != $packageVersion;
    }

    /**
     * @param string $json
     * @return mixed
     * @throws Exception
     */
    protected function _safe_json_decode($json)
    {
        $json = mb_ereg_replace(self::JSON_UNACCEPTED_CHARS, self::STRING_EMPTY, $json);
        $object = json_decode($json);
        $jsonErrorCode = json_last_error();
        if ($jsonErrorCode !== JSON_ERROR_NONE) {
            $jsonErrorMessage = function_exists(self::JSON_LAST_ERROR_MSG_FUNCTION) ? json_last_error_msg() : self::STRING_EMPTY;
            Log::getInstance()->error(sprintf(
                self::LOG_ERROR_JSON_DECODE,
                $jsonErrorCode,
                $jsonErrorMessage,
                $json
            ));
            throw new Exception(self::EX_MSG_JSON_DECODE_ERROR);
        }
        return $object;
    }

    /**
     * @return PartnerCredentials
     * @throws Exception
     */
    public function getPartnerCredentials()
    {
        Log::getInstance()->event(self::LOG_CALLED_GET_PARTNER_CREDENTIALS);
        $message = sprintf(self::EXCEPTION_METHOD_IS_NOT_APPLICABLE_MESSAGE, __METHOD__);
        Log::getInstance()->warning($message);
        throw new Exception($message);
    }

    /**
     * @throws Exception
     */
    protected function _checkResourcesStatus()
    {
        $resourcesNotReady = $this->_getApsApi()->getPackageResourcesByStatusNot(
            $this,
            array(Service_APSApi_Api::APS_READY, Service_APSApi_Api::APS_SUSPENDED),
            $this->aps->id
        );
        if (count($resourcesNotReady) > 0) {
            $message = self::ERROR_RESOURCES_NOT_READY;
            foreach ($resourcesNotReady as $resource) {
                $message .= sprintf(self::ERROR_RESOURCE_NOT_READY, $resource->aps->type, $resource->aps->id, $resource->aps->status);
            }
            throw new Exception($message);
        }
    }

    /**
     * @var PartnerCredentials $credentials
     */
    protected function _createDatacenterFromPartnerCredentials($credentials)
    {
        if (count($this->datacenters) > 0) {
            return;
        }

        if (empty($credentials->partnerLogin)) {
            return;
        }

        /**
         * @var datacenter $datacenter
         */
        $datacenter = \APS\TypeLibrary::newResourceByTypeId(self::CLOUD_BACKUP_DATACENTER_TYPE);
        $datacenter->global->set($this);
        $datacenter->id = self::DEFAULT_DATACENTER_ID;
        $datacenter->name = parse_url($credentials->baasUrl, PHP_URL_HOST);
        $datacenter->url = $credentials->baasUrl;
        $datacenter->login = $credentials->partnerLogin;
        $datacenter->password = $credentials->partnerPassword;

        $this->_apsc->registerResource($datacenter);
    }

    /**
     * @throws RestException
     */
    private function checkDiscountManagementAccess()
    {
        $headers = getallheaders();
        if (!isset($headers['APS-Identity-ID'])) {
            throw new \RuntimeException('Unable to verify request identity.');
        }

        $identityId = $headers['APS-Identity-ID'];
        $identity = $this->getProviderApsc()->getResource($identityId);

        if ($identity->memberId !== Service_OpenApi_ApiInterface::PROVIDER_ACCOUNT_ID) {
            throw new RestException(403, 'Forbidden.');
        }
    }
}
