<?php
/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

require_once(__DIR__ . '/app/bootstrap.php');

/**
 * @type("http://acronis.com/aps/cloud-backup-v1/minimal-commitment-tier/4.0")
 * @implements("http://aps-standard.org/types/core/resource/1.0")
 */
class minimalCommitmentTier extends APSResource
{
    const LOG_EVENT_METHOD_WAS_CALLED = 'Method "%s" was called for the minimum commitment tier resource "%s".';
    const LOG_EVENT_SKIP_EXCEPTION = 'Skip the exception "%s".';
    const LOG_CALLED_UPGRADE = 'Minimal commitment tier (%s) upgrade was called';

    const EXCEPTION_BAD_REQUEST_CODE = 400;

    const EXCEPTION_INVALID_VALUE_FOR_LEVEL_MESSAGE = "Invalid value for level.";
    const EXCEPTION_CANNOT_CHANGE_LEVEL_MESSAGE = "You cannot change tier level.";
    const EXCEPTION_CREATING_MINIMAL_COMMITMENT_TIER_MESSAGE = 'Creating the minimum commitment tier "%s".';
    const EXCEPTION_EMPTY_MINIMAL_COMMITMENTS_MESSAGE = 'There are no minimum commitments.';
    const EXCEPTION_INVALID_CURRENCY_MESSAGE = 'Currency "%s" is invalid.';
    const EXCEPTION_INVALID_VALUE_MESSAGE = 'Invalid value "%s" for currency "%s".';
    const EXCEPTION_NOT_UNIQUE_CURRENCY_MESSAGE = 'Currency "%s" is not unique.';

    const CURRENCY_PATTERN = '/^[A-Z]{3}$/';
    const VALUE_PATTERN = '/^(0|[1-9]\d*)(\.\d{1,2})?$/';

    const LEVEL_MINIMUM = 0;
    const LEVEL_NO_COMMITMENT = -1;
    const LEVEL_FIXED_COMMITMENT = -2;

    /**
     * @link("http://acronis.com/aps/cloud-backup-v1/globals/9.0")
     * @required
     */
    public $global;

    /**
     * @link("http://acronis.com/aps/cloud-backup-v1/tenant/9.0[]")
     */
    public $tenants;

    /**
     * @type("string")
     * @required
     */
    public $name;

    /**
     * @type("integer")
     * @required
     */
    public $level;

    /**
     * @type(MinimalCommitment[])
     */
    public $minimalCommitments = array();

    /**
     * @type("integer")
     * @readonly
     */
    public $resourceTypeId;

    public function provision()
    {
        Log::getInstance()->event(self::LOG_EVENT_METHOD_WAS_CALLED, __METHOD__, $this->aps->id);
        $this->checkLevel();
        $this->checkMinimalCommitments();
        throw new \Rest\Accepted(
            $this,
            sprintf(self::EXCEPTION_CREATING_MINIMAL_COMMITMENT_TIER_MESSAGE, $this->aps->id),
            self::ASYNC_OPERATION_TIMEOUT
        );
    }

    public function configure($resource)
    {
        if ($this->level != $resource->level) {
            throw new \Rest\RestException(
                self::EXCEPTION_BAD_REQUEST_CODE,
                self::EXCEPTION_CANNOT_CHANGE_LEVEL_MESSAGE
            );
        }
        parent::configure($resource);
        $this->checkMinimalCommitments();
    }

    public function provisionAsync()
    {
        Log::getInstance()->event(self::LOG_EVENT_METHOD_WAS_CALLED, __METHOD__, $this->aps->id);
        $this->_bindResourceType();
    }

    public function unprovision()
    {
        Log::getInstance()->event(self::LOG_EVENT_METHOD_WAS_CALLED, __METHOD__, $this->aps->id);
        $this->_unbindResourceType();
    }

    public function upgrade()
    {
        Log::getInstance()->event(sprintf(self::LOG_CALLED_UPGRADE, $this->aps->id));

        if ($this->aps->type === self::CLOUD_BACKUP_MINIMAL_COMMITMENT_TIER) {
            return;
        }

        $this->aps->type = self::CLOUD_BACKUP_MINIMAL_COMMITMENT_TIER;
        if (is_null($this->level)) {
            $this->level = 0;
        }
        $this->update();
    }

    /**
     * @return PartnerCredentials
     * @throws Exception
     */
    public function getPartnerCredentials()
    {
        Log::getInstance()->event(self::LOG_EVENT_METHOD_WAS_CALLED, __METHOD__, $this->aps->id);
        $message = sprintf(self::EXCEPTION_METHOD_IS_NOT_APPLICABLE_MESSAGE, __METHOD__);
        Log::getInstance()->warning($message);
        throw new Exception($message);
    }

    /**
     * @verb(POST)
     * @path("/unbind_resource_type")
     *
     * @param()
     */
    public function unbindResourceType()
    {
        Log::getInstance()->event(sprintf(self::LOG_EVENT_METHOD_WAS_CALLED, __METHOD__, $this->aps->id));
        $this->_unbindResourceType();
        $this->update();
        throw new Rest\RestException(200, "No Content");
    }

    private function _bindResourceType()
    {
        if (isset($this->resourceTypeId)) {
            return;
        }

        /** @var Service_OpenApi_Api $openApi */
        $openApi = $this->_getOpenApi();

        $applicationId = $openApi->apsGetApplicationId(self::PACKAGE_ID);
        $resourceId = $this->aps->id;

        $this->resourceTypeId = $openApi->addRTServiceLink($this->name, $applicationId, $resourceId, '', null);
    }

    private function _unbindResourceType()
    {
        if (!isset($this->resourceTypeId)) {
            return;
        }

        /** @var Service_OpenApi_Api $openApi */
        $openApi = $this->_getOpenApi();

        try {
            $openApi->removeResourceType($this->resourceTypeId);
        } catch (Service_OpenApi_Validator_InvalidStatusException $e) {
            $code = $e->getCode();
            if ($code != Service_OpenApi_Validator_InvalidStatusException::EX_ERR_NOT_FOUND) {
                throw $e;
            }
            Log::getInstance()->event(self::LOG_EVENT_SKIP_EXCEPTION, $e->getMessage());
        }

        $this->resourceTypeId = null;
    }

    private function checkLevel()
    {
        if (filter_var($this->level, FILTER_VALIDATE_INT, array('min_range' => self::LEVEL_FIXED_COMMITMENT)) === false) {
            throw new \Rest\RestException(
                self::EXCEPTION_BAD_REQUEST_CODE,
                self::EXCEPTION_INVALID_VALUE_FOR_LEVEL_MESSAGE
            );
        }
    }

    private function checkMinimalCommitments()
    {
        if ($this->level >= 0 && empty($this->minimalCommitments)) {
            throw new \Rest\RestException(
                self::EXCEPTION_BAD_REQUEST_CODE,
                self::EXCEPTION_EMPTY_MINIMAL_COMMITMENTS_MESSAGE
            );
        }
        if (empty($this->minimalCommitments)) {
            return;
        }
        $distinct = array();
        foreach ($this->minimalCommitments as $minimalCommitment) {
            $currency = $minimalCommitment->currency;
            if (!preg_match(self::CURRENCY_PATTERN, $currency)) {
                throw new \Rest\RestException(
                    self::EXCEPTION_BAD_REQUEST_CODE,
                    Str::format(self::EXCEPTION_INVALID_CURRENCY_MESSAGE, $currency)
                );
            }
            $value = $minimalCommitment->value;
            if (!preg_match(self::VALUE_PATTERN, $value)) {
                throw new \Rest\RestException(
                    self::EXCEPTION_BAD_REQUEST_CODE,
                    Str::format(self::EXCEPTION_INVALID_VALUE_MESSAGE, $value, $currency)
                );
            }
            if (isset($distinct[$currency])) {
                throw new \Rest\RestException(
                    self::EXCEPTION_BAD_REQUEST_CODE,
                    Str::format(self::EXCEPTION_NOT_UNIQUE_CURRENCY_MESSAGE, $currency)
                );
            }
            $distinct[$currency] = $value;
        }
    }
}
