<?php
/**
 * @Copyright © 2002-2020 Acronis International GmbH. All rights reserved
 */

namespace Acronis\Common\CodeGenerator\OfferingItems;

use Acronis\Common\CodeGenerator\Exception\OfferingItemValidationException;
use Acronis\Common\CodeGenerator\Exception\SchemaIsNotDefinedException;
use Acronis\Common\CodeGenerator\Exception\StorageCounterException;
use Acronis\Common\CodeGenerator\Validators\OfferingItemValidator;
use ArrayObject;
use Symfony\Component\Yaml\Yaml;

class OfferingItemsPreparer
{
    use StorageCountersTrait;

    private $logFileName;

    private $offeringItems;

    private $schemas;

    public function __construct($offeringItems, $schemas) {
        $this->offeringItems = $offeringItems;
        $this->schemas = $schemas;
    }

    public function setLogFileName($fileName)
    {
        $this->logFileName = $fileName;

        return $this;
    }

    /**
     * @return array
     * @throws SchemaIsNotDefinedException
     * @throws StorageCounterException
     * @throws OfferingItemValidationException
     */
    public function prepareOfferingItems()
    {
        $interpolator = new OfferingItemInterpolator();
        $predefinedStorageFields = $this->buildPredefinedStorageFields($this->offeringItems);

        $interpolate = function($initialScope, $currentScope, $additionalContext = []) use (&$interpolate, &$interpolator, $predefinedStorageFields) {
            if (isset($currentScope['_mapping_range'])) {
                $mappingRangeField = $currentScope['_mapping_range'];
                unset($currentScope['_mapping_range']);
                $items = new ArrayObject();
                foreach ($currentScope[$mappingRangeField] as $index => $rangeItemValue) {
                    // This is additional logic for obtaining correct ui_name and counter name fot storage ois that will be removed
                    // after changes in requirements
                    $oiName = $initialScope['name'];
                    $oiName = str_replace('child_storages', 'storage', $oiName);
                    if (array_key_exists('affix', $rangeItemValue) && isset($predefinedStorageFields[$oiName])) {
                        $affix = $rangeItemValue['affix'];
                        if (!isset($predefinedStorageFields[$oiName][$affix])) {
                            throw new \Exception(sprintf('No information about oi "%s" mapping to "%s", therefore some other counters of this oi are defined: %s.',
                                $oiName, $affix, json_encode($predefinedStorageFields[$oiName])
                            ));
                        }

                        $rangeItemValue = array_merge($predefinedStorageFields[$oiName][$affix], $rangeItemValue);
                        foreach (static::$optionalFieldsToAssign as $fieldName) {
                            if (!isset($predefinedStorageFields[$oiName][$affix]['options'][$fieldName])) {
                                continue;
                            }
                            $currentScope['options'][$fieldName] = $initialScope['options'][$fieldName] = $predefinedStorageFields[$oiName][$affix]['options'][$fieldName];
                        }
                    }

                    $items[] = $interpolate($initialScope, $currentScope, [
                        'range' => [
                            'value' => $rangeItemValue,
                            'index' => $index,
                        ]
                    ]);
                }

                // Return array which means that it is one item "mapped" to many
                return $items;
            }

            foreach ($currentScope as $k => $v) {
                if (is_string($k) && $k[0] === '_') {
                    $fieldName = ltrim($k, '_');

                    unset($currentScope[$k]);

                    if (is_string($v)) {
                        $template = $v;
                        $interpolatedString = $interpolator->interpolate($template, $initialScope, $additionalContext);
                        if (!empty(json_decode($interpolatedString, true))) {
                            $currentScope[$fieldName] = json_decode($interpolatedString, true);
                        } else if (!empty($interpolatedString)) {
                            // We add interpolated field when it is not empty only to be able to exclude empty fields.
                            // For example: we defined such field
                            //     _edition: '{{ oi.edition }}'
                            //     and initial data does not contain 'oi.edition'
                            // instead of adding edition: '' we will do nothing, just clear _edition property

                            $currentScope[$fieldName] = $interpolatedString;
                        }

                    }
                } else if (is_array($v)) {
                    $currentScope[$k] = $interpolate($initialScope, $v, $additionalContext);
                }
            }

            return $currentScope;
        };

        $offeringItemsList = [];
        foreach ($this->offeringItems as $item) {
            // We need to skip rows w/o counter name because it is defined by requirements
            if (empty($item['counter_name'])
                || $this->isStorageHostedCounter($item['counter_name'])
                || $this->isStorageUsageGCounter($item['counter_name'])
                || $this->isChildStoragesHostedCounter($item['counter_name'])) {
                continue;
            }

            $item += $this->getSchemaForItem($item);
            $ois = $interpolate($item, $item);
            if ($ois instanceof ArrayObject) {
                $offeringItemsList = array_merge($offeringItemsList, array_map(function($oi) {
                    return (array) $oi;
                }, $ois->getArrayCopy()));
            } else {
                $offeringItemsList[] = (array) $ois;
            }
        }

        if ($this->logFileName) {
            file_put_contents($this->logFileName, Yaml::dump($offeringItemsList));
        }

        foreach ($offeringItemsList as &$offeringItemData) {
            $validator = new OfferingItemValidator($offeringItemData);
            $validator->validate();
        }

        return $offeringItemsList;
    }

    /**
     * @param $item
     * @return
     * @throws SchemaIsNotDefinedException
     */
    private function getSchemaForItem($item)
    {
        $schemaName = $item['options']['schema'] ?: '';

        if (!array_key_exists($schemaName, $this->schemas)) {
            throw new SchemaIsNotDefinedException($schemaName);
        }

        return $this->schemas[$schemaName];
    }
}
